import datetime
import operator

from zope import component
from zope import schema
from zope.app.component.hooks import getSite
from zope.app.pagetemplate import viewpagetemplatefile
import zope.i18n.interfaces
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from z3c.form import button
from z3c.form import field
from z3c.form import form
from z3c.form import subform
import z3c.form.interfaces
import z3c.form
from plone.z3cform import z2
from plone.z3cform.widget import singlecheckboxwidget_factory
import collective.singing.interfaces
import collective.singing.message
import collective.singing.browser.subscribe
from collective.singing.channel import channel_lookup
from collective.dancing import MessageFactory as _

from plone.memoize.instance import memoize

class SubscribeForm(collective.singing.browser.subscribe.Subscribe):
    already_subscribed_message = _(
        u'Your email address is already subscribed. Click the '
        '"Send my subscription details" button below.')

class SendSecret(BrowserView):
    template = ViewPageTemplateFile('skeleton.pt')
    contents_template = ViewPageTemplateFile('sendsecret.pt')

    label = _(u"Edit existing subscriptions")

    description = _(
        u"Fill out the form below to receive an email with a link from which "
        "you can edit your subscriptions.")

    def forgot_secret_form(self):
        form = collective.singing.browser.subscribe.ForgotSecret(
            self.context, self.request)
        form.label = u''
        return form()

    def __call__(self):
        z2.switch_on(self)
        return self.template()

    def contents(self):
        return self.contents_template()

class Subscribe(BrowserView):
    template = ViewPageTemplateFile('skeleton.pt')
    contents_template = ViewPageTemplateFile('subscribe.pt')

    @property
    def send_secret_link(self):
        return _(
            u'Fill out the form below to subscribe to ${channel}. Note that '
            'this is for new subscriptions. Click here to '
            '<a href="${url}">edit your subscriptions</a>.',
            mapping={'channel': self.context.Title(),
                     'url': '%s/portal_newsletters/sendsecret.html' % getSite().absolute_url()})
    
    @property
    def label(self):
        return _(u"Subscribe to ${channel}",
                 mapping={'channel': self.context.Title()})

    def __call__(self):
        z2.switch_on(self,
                     request_layer=collective.singing.interfaces.IFormLayer)
        return self.template()

    def contents(self):
        self.subscribeform = SubscribeForm(self.context.aq_inner, self.request)
        return self.contents_template()


class Confirm(BrowserView):
    template = ViewPageTemplateFile('skeleton.pt')
    contents = ViewPageTemplateFile('status.pt')

    label = _(u"Confirming your subscription")

    successMessage = _(u"You confirmed your subscription successfully.")
    notKnownMessage = _(u"Your subscription isn't known to us.")

    def __call__(self):
        secret = self.request.form['secret']
        exists = False

        for channel in channel_lookup(only_subscribeable=True):
            subscriptions = channel.subscriptions.query(secret=secret)
            if len(subscriptions):
                exists = True
                for sub in subscriptions:
                    if sub.metadata.get('pending', False):
                        sub.metadata['pending'] = False

        if exists:
            self.status = self.successMessage
        else:
            self.status = self.notKnownMessage

        return self.template()

class Unsubscribe(BrowserView):
    template = ViewPageTemplateFile('skeleton.pt')
    contents = ViewPageTemplateFile('status.pt')

    label = _(u"Unsubscribe")

    def __call__(self):
        secret = self.request.form['secret']
        subs = self.context.aq_inner.subscriptions
        
        subscriptions = subs.query(secret=secret)
        if len(subscriptions):
            for sub in subscriptions:
                subs.remove_subscription(sub)
            self.status = _(u"You unsubscribed successfully.")
        else:
            self.status = _(u"You aren't subscribed to this channel.")

        return self.template()

class IncludeHiddenSecret(object):
    def render(self):
        html = super(IncludeHiddenSecret, self).render()
        secret = self.secret
        if secret is not None:
            index = html.find('</form>')
            html = (html[:index] +
                    '<input type="hidden" name="secret" value="%s"' % secret +
                    html[index:])
        return html

    @property
    def secret(self):
        secret = self.request.form.get('secret')
        if isinstance(secret, list):
            return secret[0]
        return secret

class SubscriptionEditForm(IncludeHiddenSecret, form.EditForm):
    template = viewpagetemplatefile.ViewPageTemplateFile('form.pt')    
    successMessage = _('Your subscription was updated.')
    removed = False
    handlers = form.EditForm.handlers

    @property
    def description(self):
        return self.context.channel.description

    @property
    def prefix(self):
        return '%s.%s.' % (
            self.context.channel.name, self.context.metadata['format'])

    @property
    def label(self):
        subscription = self.context
        value = subscription.channel.title
        if len(subscription.channel.composers) > 1:
            format = subscription.metadata['format']
            value = u"%s (%s)" % (
                value, subscription.channel.composers[format].title)
        return value

    @property
    def fields(self):
        if self.context.channel.collector is None:
            return field.Fields()
        return field.Fields(self.context.channel.collector.schema)

    def update(self):
        if len(self.fields) == 0:
            self.buttons = self.buttons.omit('apply')
        super(SubscriptionEditForm, self).update()

    @button.buttonAndHandler(_('Apply changes'), name='apply')
    def handleApply(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        changes = self.applyChanges(data)
        if changes:
            self.status = self.successMessage
        else:
            self.status = self.noChangesMessage

    @button.buttonAndHandler(_('Unsubscribe from newsletter'), name='unsubscribe')
    def handle_unsubscribe(self, action):
        secret = self.secret
        subs = self.context.channel.subscriptions
        for subscription in subs.query(secret=secret):
            subs.remove_subscription(subscription)
        self.removed = self.context
        self.status = _(u"You unsubscribed successfully.")

class SubscriptionAddForm(IncludeHiddenSecret, form.Form):
    template = viewpagetemplatefile.ViewPageTemplateFile('form.pt')
    ignoreContext = True
    
    added = None
    format = None # set by parent form
    status_already_subscribed = _(u"You are already subscribed. Fill out the form at the end of this page to be sent a link from where you can edit your subscription.")
    @property
    def description(self):
        return self.context.description
        
    @property
    def prefix(self):
        return '%s.%s.' % (self.context.name, self.format)

    @property
    def label(self):
        value = self.context.title
        if len(self.context.composers) > 1:
            value = u"%s (%s)" % (
                value, self.context.composers[self.format].title)
        return value

    @property
    def fields(self):
        if self.context.collector is not None:
            fields = field.Fields(self.context.collector.schema,
                                   prefix='collector.')
        else:
            fields = field.Fields()
        fields += field.Fields(self.context.composers[self.format].schema,
                               prefix='composer.')
        return fields

    @button.buttonAndHandler(_('Subscribe'), name='subscribe')
    def handle_subscribe(self, action):
        data, errors = self.extractData()
        
        if errors:
            self.status = form.AddForm.formErrorsMessage
            return

        extract = lambda d, prefix: dict(
            [(key.split('.', 1)[-1], value) for (key, value) in d.items()
             if key.startswith(prefix)])

        comp_data = extract(data, 'composer.')
        coll_data = extract(data, 'collector.')

        composer = self.context.composers[self.format]
        secret = collective.singing.subscribe.secret(
            self.context,
            composer,
            comp_data,
            self.request)
        secret_provided = self.secret
        if secret_provided and secret != secret_provided:
            self.status = _(
                u"There seems to be an error with the information you entered.")
            return

        metadata = dict(
            format=self.format,
            date=datetime.datetime.now(),
            pending=not secret_provided)

        # We assume here that the language of the request is the
        # desired language of the subscription:
        pl = component.queryAdapter(
            self.request, zope.i18n.interfaces.IUserPreferredLanguages)
        if pl is not None:
            metadata['languages'] = pl.getPreferredLanguages()

        # By using another method here we allow subclasses to override
        # what really happens here:
        self.add_subscription(
            self.context, secret, comp_data, coll_data, metadata,
            secret_provided)

    def add_subscription(self, context, secret, comp_data, coll_data, metadata,
                         secret_provided):
        try:
            self.added = self.context.subscriptions.add_subscription(
                self.context, secret, comp_data, coll_data, metadata)
        except ValueError, e:
            self.added = None
            self.status = self.status_already_subscribed
            return
            
        self.status = _(u"You subscribed successfully.")
        if not secret_provided:
            composer = self.context.composers[self.format]
            msg = composer.render_confirmation(self.added)
            status, status_msg = collective.singing.message.dispatch(msg)
            if status == u'sent':
                self.status = _(u"Information on how to confirm your "
                                "subscription has been sent to you.")
            else:
                # This implicitely rolls back our transaction.
                raise RuntimeError(
                    "There was an error with sending your e-mail.  Please try "
                    "again later.")


class SubscriptionSubForm(IncludeHiddenSecret, subform.EditSubForm):
    status_already_subscribed = _(u"You are already subscribed. "
                                  "Fill out the form at the end of this "
                                  "page to be sent a link from where you "
                                  "can edit your subscription.")
    status_error = _(u"There seems to be an error with the "
                     "information you entered.")
    status_sent = _(u"Information on how to confirm your "
                    "subscription has been sent to you.")
    successMessage = _('Your subscription was updated.')
    status_subscribed = _(u"You subscribed successfully.")
    status_unsubscribed = _(u"You unsubscribed successfully.")
    
    def update(self):
        super(SubscriptionSubForm, self).update()
        # set label on channel-checkbox
        # indent other widgets
        widgets = self.widgets.values()
        widgets[0].items[0]['label'] = self.label
        widgets[0].label = u''
        for widget in widgets[1:]:
            #widget.label = u""
            #widget.required = False
            widget.addClass('level-1') 


class SubscriptionAddSubForm(SubscriptionSubForm):
    template = viewpagetemplatefile.ViewPageTemplateFile('subform.pt')

    ignoreContext = True
    
    added = None
    format = None # set by parent form

    @property
    def description(self):
        return self.context.description
        
    @property
    def prefix(self):
        return '%s.%s.' % (self.context.name, self.format)

    @property
    def label(self):
        value = self.context.title
        if len(self.context.composers) > 1:
            value = u"%s (%s)" % (
                value, self.context.composers[self.format].title)
        return value

    @property
    def channel_selector(self):
        return '%s-%s' % (self.context.name,
                          self.format)

    @property
    def fields(self):
        if self.context.collector is not None:
            fields = field.Fields(self.context.collector.schema,
                                   prefix='collector.')
        else:
            fields = field.Fields()

        comp_fields = field.Fields(self.context.composers[self.format].schema,
                                   prefix='composer.')
        fields += comp_fields.omit(
            *['composer.'+f.getName() for f in self.parentForm.key_fields])

        select_field = field.Field(
            schema.Bool(
            __name__=self.channel_selector,
            title=self.label,
            default=False,
            required=False,
            ))
        select_field.widgetFactory[z3c.form.interfaces.INPUT_MODE] = (
            singlecheckboxwidget_factory)

        return field.Fields(select_field, prefix='selector.') + fields

    def update(self):
        def handleApply(self, action):
            data, errors = self.extractData()

            if not data.get(self.channel_selector):
                return

            if errors:
                self.status = form.AddForm.formErrorsMessage
                return

            extract = lambda d, prefix: dict(
                [(key.split('.', 1)[-1], value) for (key, value) in d.items()
                 if key.startswith(prefix)])

            comp_data = extract(data, 'composer.')
            coll_data = extract(data, 'collector.')

            pdata, perrors = self.parentForm.extractData()
            comp_data.update(pdata)

            composer = self.context.composers[self.format]
            secret = collective.singing.subscribe.secret(
                self.context,
                composer,
                comp_data,
                self.request)
            secret_provided = self.secret
            if secret_provided and secret != secret_provided:
                self.status = self.status_error
                return

            # Check if subscribed to other channels
            if not secret_provided and not self.parentForm.confirmation_sent:
                existing = sum(
                    [len(channel.subscriptions.query(secret=secret)) \
                     for channel in channel_lookup(only_subscribeable=True)])
                if existing:
                    self.status = self.status_already_subscribed
                    return
                
            metadata = dict(
                format=self.format,
                date=datetime.datetime.now(),
                pending=not secret_provided)

            # We assume here that the language of the request is the
            # desired language of the subscription:
            pl = component.queryAdapter(
                self.request, zope.i18n.interfaces.IUserPreferredLanguages)
            if pl is not None:
                metadata['languages'] = pl.getPreferredLanguages()

            # By using another method here we allow subclasses to override
            # what really happens here:
            self.add_subscription(
                self.context, secret, comp_data, coll_data, metadata,
                secret_provided)

        self.handlers = button.Handlers()
        self.handlers.addHandler(
            self.parentForm.buttons['apply'], handleApply)

        super(SubscriptionAddSubForm, self).update()


    def add_subscription(self, context, secret, comp_data, coll_data, metadata,
                         secret_provided):
        try:
            self.added = self.context.subscriptions.add_subscription(
                self.context, secret, comp_data, coll_data, metadata)
        except ValueError, e:
            self.added = None
            self.status = self.status_already_subscribed
            return

        self.status = self.status_subscribed

        if not secret_provided:
            self.parentForm.send_confirmation(self.context, self.format, self.added)

class SubscriptionEditSubForm(SubscriptionSubForm):
    template = viewpagetemplatefile.ViewPageTemplateFile('subform.pt')    
    removed = False
    handlers = form.EditForm.handlers


    @property
    def description(self):
        return self.context.channel.description

    @property
    def prefix(self):
        return '%s.%s.' % (
            self.context.channel.name, self.context.metadata['format'])

    @property
    def label(self):
        subscription = self.context
        value = subscription.channel.title
        if len(subscription.channel.composers) > 1:
            format = subscription.metadata['format']
            value = u"%s (%s)" % (
                value, subscription.channel.composers[format].title)
        return value

    @property
    def channel_selector(self):
        return '%s-%s' % (self.context.channel.name,
                          self.context.metadata['format'])

    @property
    def fields(self):
        select_field = field.Field(
            schema.Bool(
            __name__=self.channel_selector,
            title=self.label,
            default=True,
            required=False
            ))
        select_field.widgetFactory[z3c.form.interfaces.INPUT_MODE] = (
            singlecheckboxwidget_factory)
        
        fields = field.Fields(select_field, prefix='selector.')

        if self.context.channel.collector is not None:
            fields += field.Fields(
                self.context.channel.collector.schema,
                prefix='collector.') 
        return fields
    
    def update(self):
        def handleApply(self, action):
            data, errors = self.extractData()

            if not data.get(self.channel_selector):
                self.unsubscribe()
                return
            if errors:
                self.status = self.formErrorsMessage
                return

            content = self.getContent()
            del data[self.channel_selector] 

            changes = form.applyChanges(self, content, data)
            if changes:
                zope.event.notify(
                    zope.lifecycleevent.ObjectModifiedEvent(content))
                self.status = self.successMessage
            else:
                self.status = self.noChangesMessage

        self.handlers = button.Handlers()
        self.handlers.addHandler(
            self.parentForm.buttons['apply'], handleApply)

        super(SubscriptionEditSubForm, self).update()


    def unsubscribe(self):
        secret = self.secret
        format = self.context.metadata['format']
        subs = self.context.channel.subscriptions
        for subscription in subs.query(secret=secret, format=format):
            subs.remove_subscription(subscription)
        self.removed = self.context
        self.status = self.status_unsubscribed

class PrettySubscriptionsForm(IncludeHiddenSecret, form.EditForm):

    template = viewpagetemplatefile.ViewPageTemplateFile(
        'prettysubscriptionsform.pt')
    ignoreContext = True
    ignoreRequest = True
    confirmation_sent = False
    status_message = None
    
    def __init__(self, context, request, subs, channels):
        super(PrettySubscriptionsForm, self).__init__(context, request)
        self.subs = subs
        self.channels = channels
        self.key_fields = []
        composers = reduce(
            lambda x,y:x+y,
            [c.composers.values() for c in channel_lookup(only_subscribeable=True)])
        for composer in composers:
            for name in composer.schema.names():
                f = composer.schema.get(name)
                if f and \
                       collective.singing.interfaces.ISubscriptionKey.providedBy(f):
                    if f not in self.key_fields:
                        self.key_fields.append(f)
        self.confirmation_sent = False

    def status(self):
        if self.status_message:
            return self.status_message
        stati = [form.status for form in self.forms]

        for s in [SubscriptionSubForm.status_already_subscribed,
                  SubscriptionSubForm.status_sent,
                  SubscriptionSubForm.successMessage]:
            if s in stati:
                return s

        for s in [SubscriptionSubForm.status_subscribed,
                  SubscriptionSubForm.status_unsubscribed]:
            if s in stati:
                return SubscriptionSubForm.successMessage

        return ''

    @property
    def addforms(self):
        return [f for f in self.subscription_addforms if not f.added]

    @property
    def editforms(self):
        return [f for f in self.subscription_editforms if not f.removed]

    @property
    def forms(self):
        return sorted(
            self.addforms + self.editforms,
            key=operator.attrgetter('label'))

    @property
    def fields(self):
        fields = field.Fields()
        for kf in self.key_fields:
            f = field.Field(kf)
            if self.subs:
                kf.default = self.subs[0].composer_data[kf.getName()]
            fields += field.Fields(f)
        return fields

    def update(self):
        super(PrettySubscriptionsForm, self).update()

        if self.subs: #existing subscriptions
            for f_name in [f.getName() for f in self.key_fields]:
                widget = self.widgets.get(f_name)

        # Let's set convert any 'pending' subscriptions to non-pending:
        for sub in self.subs:
            if sub.metadata.get('pending'):
                sub.metadata['pending'] = False

        # Assemble the list of edit forms
        self.subscription_editforms = [
            SubscriptionEditSubForm(s, self.request, self) for s in self.subs]

        # Assemble the list of add forms
        self.subscription_addforms = []
        for format, channel in self.channels:
            addform = SubscriptionAddSubForm(channel, self.request, self)
            addform.format = format
            self.subscription_addforms.append(addform)

        # The edit forms might have deleted a subscription.  We'll
        # take care of this while updating them:
        for form in self.subscription_editforms:
            form.update()
            if form.removed:
                subscription = form.context
                name = subscription.channel.name
                addform = SubscriptionAddSubForm(
                    subscription.channel, self.request, self)
                addform.format = subscription.metadata['format']
                addform.ignoreRequest = True
                addform.update()
                self.subscription_addforms.append(addform)
                addform.status = form.status
            #elif form.status != form.noChangesMessage:
            #    self.status = form.status
        
        # Let's update the add forms now.  One of them may have added
        # a subscription:
        for form in self.subscription_addforms:
            form.update()
            subscription = form.added
            if subscription is not None:
                editform = SubscriptionEditSubForm(
                    subscription, self.request, self)
                editform.update()
                self.subscription_editforms.append(editform)
                editform.status = form.status#_(u"You subscribed successfully.")

    @button.buttonAndHandler(_('Apply'), name='apply')
    def handle_apply(self, action):
        # All the action happens in the subforms ;-)
        # All we do here is check that key_fields were not altered
        if self.subs:
            data, errors = self.extractData()
            key_defaults = dict(
                [(f.getName(), f.default) for f in self.key_fields])
            for key, value in data.items():
                if key_defaults[key] != value:
                    self.status_message = SubscriptionSubForm.status_error

    def send_confirmation(self, channel, format, subscription):
        if not self.confirmation_sent:
            composer = channel.composers[format]
            msg = composer.render_confirmation(subscription)
            status, status_msg = collective.singing.message.dispatch(msg)
            if status == u'sent':
                self.status_message = _(u"Information on how to confirm your "
                                        "subscription has been sent to you.")
                self.confirmation_sent = True
            else:
                # This implicitely rolls back our transaction.
                raise RuntimeError(
                    "There was an error with sending your e-mail.  Please try "
                    "again later.")

class Subscriptions(BrowserView):
    __call__ = ViewPageTemplateFile('skeleton.pt')
    contents_template = ViewPageTemplateFile('subscriptions.pt')
    single_form_template = ViewPageTemplateFile('prettysubscriptions.pt')
 
    label = _(u"My subscriptions")
    status = u""

    def forgot_secret_form(self):
        form = collective.singing.browser.subscribe.ForgotSecret(
            self.context, self.request)
        form.label = u''
        return form()

    @property
    def newsletters(self):
        return getSite().portal_newsletters
    
    @property
    def single_form_subscriptions(self):
        return self.newsletters.get('use_single_form_subscriptions_page', '')

    @property
    def addforms(self):
        return [f for f in self.subscription_addforms if not f.added]

    @property
    def editforms(self):
        return [f for f in self.subscription_editforms if not f.removed]

    @property
    def secret(self):
        secret = self.request.form.get('secret')
        if isinstance(secret, list):
            return secret[0]
        return secret

    def contents(self):
        z2.switch_on(self,
                     request_layer=collective.singing.interfaces.IFormLayer)
        subscriptions, channels = self._subscriptions_and_channels(self.secret)
        
        if self.single_form_subscriptions:
            self.form = PrettySubscriptionsForm(self.context, self.request, 
                                                subscriptions, channels)
            self.form.update()
            if len(self.form.key_fields) == 1:
                # Revert to old style form if not all
                # composers have the same ISubscriptionKey.
                return self.single_form_template()

        # Let's set convert any 'pending' subscriptions to non-pending:
        for sub in subscriptions:
            if sub.metadata.get('pending'):
                sub.metadata['pending'] = False

        # Assemble the list of edit forms
        self.subscription_editforms = [
            SubscriptionEditForm(s, self.request) for s in subscriptions]

        # Assemble the list of add forms
        self.subscription_addforms = []
        for format, channel in channels:
            addform = SubscriptionAddForm(channel, self.request)
            addform.format = format
            self.subscription_addforms.append(addform)

        # The edit forms might have deleted a subscription.  We'll
        # take care of this while updating them:
        for form in self.subscription_editforms:
            form.update()
            if form.removed:
                subscription = form.context
                name = subscription.channel.name
                addform = SubscriptionAddForm(
                    subscription.channel, self.request)
                addform.format = subscription.metadata['format']
                addform.update()
                self.subscription_addforms.append(addform)
                addform.status = form.status
            elif form.status != form.noChangesMessage:
                self.status = form.status

        # Let's update the add forms now.  One of them may have added
        # a subscription:
        for form in self.subscription_addforms:
            form.update()
            subscription = form.added
            if subscription is not None:
                editform = SubscriptionEditForm(
                    subscription, self.request)
                editform.update()
                self.subscription_editforms.append(editform)
                editform.status = form.status#_(u"You subscribed successfully.")

        return self.contents_template()

    def _subscriptions_and_channels(self, secret):
        subscriptions = []
        channels_and_formats = []

        for channel in channel_lookup(only_subscribeable=True):
            channel_subs = channel.subscriptions

            subscribed_formats = []
            if self.secret is not None:
                for s in channel_subs.query(secret=self.secret):
                    subscriptions.append(s)
                    subscribed_formats.append(s.metadata['format'])

            for format in channel.composers.keys():
                if format not in subscribed_formats:
                    channels_and_formats.append((format, channel))

        return subscriptions, channels_and_formats


#############
from zope import schema
from zope.schema.vocabulary import SimpleVocabulary
from collective.dancing.composer import HTMLComposer

class MyHTMLComposer(HTMLComposer):
    title = u'Hypertext E-Mail with selectable font-size'
    class schema(HTMLComposer.schema):
        font_size = schema.Choice(
            title=u"Font size",
            vocabulary=SimpleVocabulary.fromValues([8, 12, 16]))



import copy_reg
import persistent
import zc.queue
import logging

import collective.singing.channel
import collective.singing.subscribe
import collective.dancing.composer
from collective.singing.queue import CompositeQueue

logger = logging.getLogger('collective.dancing')
safe_reconstructor = copy_reg._reconstructor

def _reconstructor(cls, base, state):
    if issubclass(cls, collective.dancing.composer.HTMLComposer):
        obj = persistent.Persistent.__new__(cls, state)
        base.__init__(obj, state)
        return obj
    else:
        return safe_reconstructor(cls, base, state)

def fix_legacy_htmlcomposers(tool):
    """collective.dancing.composer.HTMLComposer used to derive from
    object.  We need to update all pickles of composers with this
    little brute force function.
    """
    site = tool.aq_parent
    copy_reg._reconstructor = _reconstructor
    try:
        if 'portal_newsletters' in site.objectIds():
            for channel in site['portal_newsletters'].channels.values():
                channel.composers = channel.composers
    finally:
        copy_reg._reconstructor = _reconstructor

def upgrade_to_compositequeue(tool):
    """collective.singing.message.MessageQueues used to store messages
    in zc.queue.Queue objects, which are inefficient for large queues.
    This upgrade modifies existing instances of MessageQueues to use
    zc.queue.CompositeQueue instead."""
    for channel in collective.singing.channel.channel_lookup():
        for key in channel.queue.keys():
            new = zc.queue.CompositeQueue()
            for item in channel.queue[key]:
                new.put(item)
            channel.queue[key] = new

def reindex_subscriptions(tool):
    """The fulltext index for subscriptions wasn't populated before
    version 0.7.2.  This upgrade step will reindex all subscriptions
    that it can find.
    """
    for channel in collective.singing.channel.channel_lookup():
        for sub in channel.subscriptions.values():
            collective.singing.subscribe._catalog_subscription(sub)

def upgrade_to_singing_compositequeue(tool):
    """collective.singing.message.MessageQueues used to store messages
    in zc.queue.CompositeQueue objects, which are inefficient when counting
    items in a large queue, because it needs to access many persistent sub-
    queues.
    This upgrade modifies existing instances of MessageQueues to use
    collective.singing.message.CompositeQueue instead."""
    for channel in collective.singing.channel.channel_lookup():
        logger.info("Updating queues in %s to singing CompositeQueue" % channel.name)
        for key in channel.queue.keys():
            if isinstance(channel.queue[key], CompositeQueue):
                continue
            new = CompositeQueue()
            for item in channel.queue[key]:
                new.put(item)
            channel.queue[key] = new
        logger.info("Updated queues in %s to singing CompositeQueue" % channel.name)

def upgrade_scheduled_sends(tool):
    """ Upgrade already scheduled items in Timed Schedulers
    to reflect code changes in 0.8.9.
    I.e. scheduler.items is now a list of (when, content, override_vars)
    instead of the previous (when, content)"""
    for channel in collective.singing.channel.channel_lookup():
        if isinstance(channel.scheduler,
                      collective.singing.scheduler.TimedScheduler):
            items = persistent.list.PersistentList()
            for item in channel.scheduler.items:
                try:
                    when, content = item
                    items.append((when, content, {}))
                    logger.info('Upgrading scheduled item "%s" for channel "%s"' \
                                % (item, channel.Title()))
                except ValueError:
                    items.append(item)
            channel.scheduler.items = items



import datetime
from zope import component
from zope import interface
from zope import schema
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope.app.pagetemplate import viewpagetemplatefile

from z3c.form import button
from z3c.form import field

from plone.app.z3cform import layout
from plone.z3cform import z2
from plone.z3cform.crud import crud
from collective.singing import interfaces
from collective.singing.channel import channel_lookup

from collective.dancing.browser import controlpanel
from collective.dancing import utils
from collective.dancing import MessageFactory as _

class IQueueStatistics(interface.Interface):
    title = schema.TextLine(title=_(u"Title"))

    messages_sent = schema.Int(
        title=_(u"Total messages sent"),
        description=_(u"Total number of messages sent successsfully"),
        )

    new_messages = schema.Int(
        title=_(u"New messages"),
        description=_(u"Messages waiting to be sent"),
        )
    new_last_modified = schema.Datetime(title=_(u"Last Change"))

    sent_messages = schema.Int(
        title=_(u"Sent messages"),
        description=_(u"Messages successfully sent"),
        )
    sent_last_modified = schema.Datetime(title=_(u"Last Change"))

    error_messages = schema.Int(
        title=_(u"Failed messages"),
        description=_(u"Messages with errors"),
        )
    error_last_modified = schema.Datetime(title=_(u"Last Change"))

    retry_messages = schema.Int(
        title=_(u"Retry messages"),
        description=_(u"Messages with errors in the retry queue"),
        )
    retry_last_modified = schema.Datetime(title=_(u"Last Change"))

class ChannelStatistics(object):
    interface.implements(IQueueStatistics)
    component.adapts(interfaces.IChannel)

    def __init__(self, channel):
        self.channel = channel
        self.title = channel.title

        self.messages_sent = channel.queue.messages_sent
        for status in interfaces.MESSAGE_STATES:
            size, date = self._get_queue_info(channel.queue, status)
            setattr(self, '%s_messages' % status, size)
            setattr(self, '%s_last_modified' % status, date)

    @staticmethod
    def _get_queue_info(queue, status):
        msg_count = len(queue[status])
        if msg_count:
            last_modified = queue[status][-1].status_changed
        else:
            last_modified = None
        return msg_count, last_modified

class EditForm(crud.EditForm):
    jobs_template = viewpagetemplatefile.ViewPageTemplateFile('jobs.pt')

    def render(self):
        # In addition to rendering the usual crud table, we'll add a
        # table with information about jobs at the top:
        html = u''
        if self.pending_jobs() or self.finished_jobs():
            html = self.jobs_template()

        return html + super(EditForm, self).render()

    @button.buttonAndHandler(_('Remove old messages'), name='removeold')
    def handle_clear(self, action):
        self.status = _(u"Please select items to remove.")
        selected = self.selected_items()
        if selected:
            self.status = _(u"Successfully removed old messages from channels.")
            for id, stats in selected:
                stats.channel.queue.clear()

    @button.buttonAndHandler(_('Clear queued messages'), name='clearnew')
    def handle_clearnew(self, action):
        self.status = _(u"Please select items with new messages to clear.")
        selected = self.selected_items()
        if selected:
            self.status = _(u"Successfully cleared queued messages in channels.")
            for id, stats in selected:
                stats.channel.queue.clear(queue_names=('new','retry'))

    @button.buttonAndHandler(_('Send queued messages now'), name='send')
    def handle_send(self, action):
        self.status = _(u"Please select which channel's queued e-mails"
                        " you'd like to send.")
        selected = self.selected_items()
        if selected:
            sent, failed = 0, 0
            for id, stats in selected:
                s, f = stats.channel.queue.dispatch()
                sent += s
                failed += f
            self.status = _(u"${sent} message(s) sent, ${failed} failure(s).",
                            mapping=dict(sent=sent, failed=failed))

    @button.buttonAndHandler(_('Process jobs'), name='process_jobs',
                             condition=lambda form: form.pending_jobs())
    def handle_process_jobs(self, action):
        queue = utils.get_queue()
        num = queue.process()
        finished = queue.finished[-num:]
        if len(finished) == 1:
            self.status = finished[0].value
        else:
            self.status = _(u"All pending jobs processed")

    @button.buttonAndHandler(_('Clear finished jobs'), name='clear_jobs',
                             condition=lambda form: form.finished_jobs())
    def handle_clear_jobs(self, action):
        queue = utils.get_queue()
        while queue.finished:
            queue.finished.pop()
        self.status = _("All finished jobs cleared")

    def pending_jobs(self):
        return utils.get_queue().pending

    def finished_jobs(self):
        return utils.get_queue().finished

class StatsForm(crud.CrudForm):
    """View list of queues and last modification dates for all
    available channels.
    """
    addform_factory = crud.NullForm
    editform_factory = EditForm
    view_schema = IQueueStatistics

    def get_items(self):
        return [(channel.name, ChannelStatistics(channel))
                for channel in channel_lookup()]

    def remove(self, (id, item)):
        for name, stats in self.get_items():
            if name == id:
                stats.channel.queue.clear()
                self.status = _(u"All sent and failed messages deleted")
                return
        else:
            raise KeyError(id)

StatsView = layout.wrap_form(
    StatsForm,
    index=ViewPageTemplateFile('controlpanel.pt'),
    label = _(u"Newsletter statistics"),
    back_link = controlpanel.back_to_controlpanel)



# We patch our friend, the ATTopic to allow overriding *all* of the
# query that it builds by providing keyword arguments.

from Products.ATContentTypes.content.topic import *

def queryCatalog(self, REQUEST=None, batch=False, b_size=None,
                                                full_objects=False, **kw):
    """Invoke the catalog using our criteria to augment any passed
        in query before calling the catalog.
    """
    if REQUEST is None:
        REQUEST = getattr(self, 'REQUEST', {})
    b_start = REQUEST.get('b_start', 0)

    pcatalog = getToolByName(self, 'portal_catalog')
    mt = getToolByName(self, 'portal_membership')
    related = [ i for i in self.getRelatedItems() \
                    if mt.checkPermission(View, i) ]
    if not full_objects:
        related = [ pcatalog(path='/'.join(r.getPhysicalPath()))[0] 
                    for r in related]
    related=LazyCat([related])

    limit = self.getLimitNumber()
    max_items = self.getItemCount()
    # Batch based on limit size if b_szie is unspecified
    if max_items and b_size is None:
        b_size = int(max_items)
    else:
        b_size = b_size or 20

    q = self.buildQuery()
    if q is None:
        results=LazyCat([[]])
    else:
        # Allow parameters to further limit existing criterias -- for real!
        q.update(kw)

        if not batch and limit and max_items and self.hasSortCriterion():
            # Sort limit helps Zope 2.6.1+ to do a faster query
            # sorting when sort is involved
            # See: http://zope.org/Members/Caseman/ZCatalog_for_2.6.1
            q.setdefault('sort_limit', max_items)
        __traceback_info__ = (self, q)
        results = pcatalog.searchResults(REQUEST, **q)

    if limit and not batch:
        if full_objects:
            return related[:max_items] + \
                   [b.getObject() for b in results[:max_items-len(related)]]
        return related[:max_items] + results[:max_items-len(related)]
    elif full_objects:
        results = related + LazyCat([[b.getObject() for b in results]])
    else:
        results = related + results
    if batch:
        batch = Batch(results, b_size, int(b_start), orphan=0)
        return batch
    return results

ATTopic.queryCatalog = queryCatalog



import five.intid.site

def importVarious(context):
    """Import various settings.

    Provisional handler that does initialization that is not yet taken
    care of by other handlers.
    """
    # Only run step if a flag file is present
    if context.readDataFile('dancing_various.txt') is None:
        return
    site = context.getSite()
    five.intid.site.add_intids(site)



def beforeUninstall(portal, reinstall, product, cascade):
    """Don't remove our utilities on reinstall, thank you!"""
    if reinstall:
        return '', filter(lambda c: c not in ('utilities', 'adapters'), cascade)
    else:
        return '', cascade



import cStringIO
import datetime
from DateTime import DateTime
import sys
import sets
import csv
from zope import interface
from zope import component
from zope import schema
import zope.interface
import zope.schema.interfaces
import zope.schema.vocabulary
import zope.i18n
from z3c.form import field, form
import z3c.form.interfaces
import z3c.form.datamanager
import z3c.form.term
import z3c.form.button
import OFS.SimpleItem
from Globals import DevelopmentMode
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.CMFCore.interfaces import IPropertiesTool
import Products.CMFPlone.utils
from collective.singing.interfaces import IChannel, IFormLayer, ICollectorSchema
from plone.z3cform import z2
from plone.z3cform.crud import crud
from plone.app.z3cform import wysiwyg
import collective.singing.scheduler
import collective.singing.subscribe
import collective.singing.channel
from zope.app.pagetemplate import viewpagetemplatefile
from collective.dancing import MessageFactory as _
from collective.dancing import collector
from collective.dancing import utils
from collective.dancing import channel
from collective.dancing.composer import check_email
from collective.dancing.browser import controlpanel
from collective.dancing.browser.interfaces import ISendAndPreviewForm

def simpleitem_wrap(klass, name):
    class SimpleItemWrapper(klass, OFS.SimpleItem.SimpleItem):
        __doc__ = OFS.SimpleItem.SimpleItem.__doc__
        id = name
        def Title(self):
            return klass.title

    klassname = klass.__name__
    SimpleItemWrapper.__name__ = klassname
    module = sys.modules[__name__]
    assert not hasattr (module, klassname), "%r already a name in this module."
    setattr(module, klassname, SimpleItemWrapper)
    return SimpleItemWrapper

schedulers = [
    simpleitem_wrap(klass, 'scheduler')
    for klass in collective.singing.scheduler.schedulers]
    
csv_delimiter = ","

class FactoryChoice(schema.Choice):
    def _validate(self, value):
        if self._init_field:
            return
        super(schema.Choice, self)._validate(value)

        # We'll skip validating against the vocabulary

def scheduler_vocabulary(context):
    terms = []
    for factory in schedulers:
        terms.append(
            zope.schema.vocabulary.SimpleTerm(
                value=factory(),
                token='%s.%s' % (factory.__module__, factory.__name__),
                title=factory.title))
    return utils.LaxVocabulary(terms)
zope.interface.alsoProvides(scheduler_vocabulary,
                            zope.schema.interfaces.IVocabularyFactory)

class ChannelEditForm(crud.EditForm):
    def _update_subforms(self):
        self.subforms = []
        for channel in collective.singing.channel.channel_lookup():
            subform = crud.EditSubForm(self, self.request)
            subform.content = channel
            subform.content_id = channel.name
            subform.update()
            self.subforms.append(subform)

class ManageChannelsForm(crud.CrudForm):
    """Crud form for channels."""
    
    description = _("Add or edit channels that will use collectors to "
                    "gather and email specific sets of information from "
                    "your site, to subscribed email addresses, at scheduled "
                    "times.")
    
    editform_factory = ChannelEditForm

    @property
    def add_schema(self):
        if len(channel.channels) > 1:
            return self.update_schema + field.Fields(
                schema.Choice(
                __name__='factory',
                title=_(u"Type"),
                vocabulary=zope.schema.vocabulary.SimpleVocabulary(
                    [zope.schema.vocabulary.SimpleTerm(value=c, title=c.type_name)
                     for c in channel.channels])
                ))
        return self.update_schema

    @property
    def update_schema(self):
        fields = field.Fields(IChannel).select('title')

        collector = schema.Choice(
            __name__='collector',
            title=IChannel['collector'].title,
            required=False,
            vocabulary='Collector Vocabulary')

        scheduler = FactoryChoice(
            __name__='scheduler',
            title=IChannel['scheduler'].title,
            required=False,
            vocabulary='Scheduler Vocabulary')

        fields += field.Fields(collector, scheduler)

        return fields

    @property
    def view_schema(self):
        return self.update_schema.copy()
    
    def get_items(self):
        return collective.singing.channel.channel_lookup()
    
    def add(self, data):
        name = Products.CMFPlone.utils.normalizeString(
            data['title'].encode('utf-8'), encoding='utf-8')
        factory = data.get('factory', None) or channel.channels[0]
        self.context[name] = factory(
            name, data['title'],
            collector=data['collector'],
            scheduler=data['scheduler'])
        return self.context[name]

    def remove(self, (id, item)):
        self.context.manage_delObjects([id])

    def link(self, item, field):
       if field == 'title':
           return item.absolute_url()
       elif field == 'collector' and item.collector is not None:
           collector_id = item.collector.getId()
           collector = getattr(self.context.aq_inner.aq_parent.collectors,
                               collector_id)
           return collector.absolute_url()
       elif field == 'scheduler':
           if item.scheduler is not None:
               return item.scheduler.absolute_url()
            
class ChannelAdministrationView(BrowserView):
    __call__ = ViewPageTemplateFile('controlpanel.pt')
    
    label = _(u'Channel administration')
    back_link = controlpanel.back_to_controlpanel

    def contents(self):
        # A call to 'switch_on' is required before we can render z3c.forms.
        z2.switch_on(self)
        return ManageChannelsForm(self.context.channels, self.request)()

class SubscriptionsSearchForm(z3c.form.form.Form):
    prefix = 'search.'
    ignoreContext = True

    fields = field.Fields(
        schema.TextLine(
        __name__='fulltext',
        title=_(u"Search subscribers"),
        ))

    @z3c.form.button.buttonAndHandler(_('Search'), name='search')
    def handle_search(self, action):
        pass

class ManageSubscriptionsFormEdit(crud.EditForm):
    def update(self):
        super(ManageSubscriptionsFormEdit, self).update()
        self.search = SubscriptionsSearchForm(self.context, self.request)
        self.search.update()

    def render(self):
        table = super(ManageSubscriptionsFormEdit, self).render()
        name = self.search.widgets['fulltext'].name
        search = self.request.form.get(name)
        if search and table.strip():
            idx = table.find('</form>')
            hidden = ('<input type="hidden" name="%s" value="%s" />' %
                      (name, search))
            table = table[:idx] + hidden + table[idx:]
        return ('<div id="subscriber-search">%s</div>' % self.search.render() +
                table)

class ManageSubscriptionsForm(crud.CrudForm):
    """Crud form for subscriptions.
    """
    # These are set by the SubscriptionsAdministrationView
    format = None 
    composer = None

    description = _(u"Manage or add subscriptions.")

    editform_factory = ManageSubscriptionsFormEdit
    
    @property
    def batch_size(self):
        if self._fulltext_query():
            # We don't support batching when we search
            return 0
        else:
            return 30

    @property
    def prefix(self):
        return self.format

    def _composer_fields(self):
        return field.Fields(self.composer.schema)

    def _collector_fields(self):
        if self.context.collector is not None:
            return field.Fields(self.context.collector.schema)
        return field.Fields()

    def _fulltext_query(self):
        return self.request.form.get('search.widgets.fulltext')

    @property
    def update_schema(self):
        fields = self._composer_fields()
        fields += self._collector_fields()
        return fields

    def get_items(self):
        items = []
        
        query = dict(format=self.format)
        search = self._fulltext_query()
        if search:
            query['fulltext'] = search

        for subscription in self.context.subscriptions.query(**query):
            if subscription.metadata['format'] == self.format:
                items.append((str(subscription.secret), subscription))
        return items

    def add(self, data):
        secret = collective.singing.subscribe.secret(
            self.context, self.composer, data, self.request)

        composer_data = dict(
            [(name, value) for (name, value) in data.items()
             if name in self._composer_fields()])

        collector_data = dict(
            [(name, value) for (name, value) in data.items()
             if name in self._collector_fields()])

        metadata = dict(format=self.format,
                        date=datetime.datetime.now())

        try:
            return self.context.subscriptions.add_subscription(
                self.context, secret, composer_data, collector_data, metadata)
        except ValueError, e:
            raise schema.ValidationError(e.args[0])

    def remove(self, (secret, item)):
        subs = self.context.subscriptions.query(secret=secret, 
                                                format=item.metadata['format'])
        for subscription in subs:
            self.context.subscriptions.remove_subscription(subscription)


class SubscriptionChoiceFieldDataManager(z3c.form.datamanager.AttributeField):
    # This nasty hack allows us to have the default IDataManager to
    # use a different schema for adapting the context.  This is
    # necessary because the schema that
    # ``collector.SmartFolderCollector.schema`` produces is a
    # dynamically generated interface.
    #
    # ``collector.SmartFolderCollector.schema`` should rather produce
    # an interface with fields that already have the right interface
    # to adapt to as their ``interface`` attribute.
    component.adapts(
        collective.singing.subscribe.SimpleSubscription,
        zope.schema.interfaces.IField)

    def __init__(self, context, field):
        super(SubscriptionChoiceFieldDataManager, self).__init__(context, field)
        if self.field.interface is not None:
            if issubclass(self.field.interface, ICollectorSchema):
                self.field.interface = ICollectorSchema

class ChannelPreviewForm(z3c.form.form.Form):
    """Channel preview form.

    Currently only allows an in-browser preview.
    """

    interface.implements(ISendAndPreviewForm)

    template = viewpagetemplatefile.ViewPageTemplateFile('form.pt')
    description = _(u"See an in-browser preview of the newsletter.")

    fields = z3c.form.field.Fields(ISendAndPreviewForm).select(
        'include_collector_items')

    include_collector_items = True
    
    def getContent(self):
        return self
    
    @z3c.form.button.buttonAndHandler(_(u"Generate"), name='preview')
    def handle_preview(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = form.EditForm.formErrorsMessage
            return

        collector_items = int(bool(data['include_collector_items']))
        
        return self.request.response.redirect(
            self.context.absolute_url()+\
            '/preview-newsletter.html?include_collector_items=%d' % \
            collector_items)

class EditChannelForm(z3c.form.form.EditForm):
    """Channel edit form.

    As opposed to the crud form, this allows editing of all channel
    settings.

    Actions are also provided to preview and send the newsletter.
    """

    template = viewpagetemplatefile.ViewPageTemplateFile('form.pt')

    description = _(u"Edit the properties of the channel.")
    
    @property
    def fields(self):
        fields = z3c.form.field.Fields(IChannel).select('title')

        collector = schema.Choice(
            __name__='collector',
            title=IChannel['collector'].title,
            required=False,
            vocabulary='Collector Vocabulary')

        scheduler = FactoryChoice(
            __name__='scheduler',
            title=IChannel['scheduler'].title,
            required=False,
            vocabulary='Scheduler Vocabulary')
        
        fields += field.Fields(collector, scheduler)
        fields += field.Fields(IChannel).select('description')
        fields['description'].widgetFactory[
            z3c.form.interfaces.INPUT_MODE] = wysiwyg.WysiwygFieldWidget

        return fields


def parseSubscriberCSVFile(subscriberdata, composer, 
                           header_row_present=False,
                           delimiter=csv_delimiter):
    """parses file containing subscriber data

    returns list of dictionaries with subscriber data according to composer"""
    properties = component.getUtility(IPropertiesTool)
    charset = properties.site_properties.getProperty('default_charset', 
                                                     'utf-8').upper()    
    try:
        data = cStringIO.StringIO(subscriberdata)
        reader = csv.reader(data, delimiter=str(delimiter))
        subscriberslist = []
        errorcandidates = []
        for index, parsedline in enumerate(reader):
            if index == 0:
                if header_row_present:
                    fields = parsedline
                    continue
                else:
                    fields = field.Fields(composer.schema).keys()
            if len(parsedline)<len(fields):
                pass
            else:
                try:
                    subscriber = dict(zip(fields,\
                       map(lambda x:x.decode(charset), parsedline)))
                    check_email(subscriber['email'])
                except:
                    errorcandidates.append(subscriber['email'])
                else:
                    subscriberslist.append(subscriber)
        return subscriberslist, errorcandidates
    except Exception, e:
        return _(u"Error importing subscriber file. %s" % e), []

class ExportCSV(BrowserView):

    def __call__(self):
        properties = component.getUtility(IPropertiesTool)
        charset = properties.site_properties.getProperty('default_charset', 
                                                         'utf-8').upper()
        self.request.response.setHeader('Content-Type', 
                                        'text/csv; charset=%s' % charset)
        self.request.response.setHeader('Content-disposition',
            'attachment; filename=subscribers_%s_%s.csv' % (self.context.id, 
                                      datetime.date.today().strftime("%Y%m%d")))
        res = cStringIO.StringIO()
        writer = csv.writer(res, dialect=csv.excel, delimiter=csv_delimiter)
        for format in self.context.composers.keys():
            for subscription in tuple(self.context.subscriptions.query(format=format)):
                row = []
                for item in field.Fields(self.context.composers[format].schema).keys():
                    v = subscription.composer_data.get(item) or ''
                    row.append(v.encode(charset))

                writer.writerow(row)
        return res.getvalue()
                                
class UploadForm(crud.AddForm):
    label = _(u"Upload")
    
    @property
    def fields(self):
        subscriberdata = schema.Bytes(
            __name__ = 'subscriberdata',
            title=_(u"Subscribers"),
            description=_(u"Upload a CSV file with a list of subscribers here. "
                          u"Subscribers already present in the database will "
                          u"be overwritten. Each line should contain: "
                          u"${columns}.",
                          mapping=dict(columns=';'.join(field.Fields(
                                         self.context.composer.schema).keys())))
        ) 
        remove = schema.Bool(__name__ = 'removenonexisting',
            title=_(u"Remove non-existing"),
            description=_(u"If selected and a subscriber exists in the "
                          u"database, but is not part of the uploaded file, "
                          u"the subscriber will be removed from the list."),
            default = False
        ) 
        header_row_present = schema.Bool(__name__ = 'header_row_present',
            title=_(u"CSV contains a header row"),
            description=_(u"Select this if you want to use the csv "
                          u"header row to designate document variables."
                          u"The header row must contain one 'email' field."),
            default = False
        ) 
        csv_delimiter = schema.TextLine(__name__ = 'csv_delimiter',
            title=_(u"CSV delimiter"),
            description=_(u"The delimiter in your CSV file. "
                          u"(Usually ',' or ';', but no quotes are necessary.)"),
            default = u','
        ) 

        return field.Fields(subscriberdata, remove, 
                            header_row_present, csv_delimiter)
   
    @property
    def mychannel(self):
        return self.context.context
        
    def _releaseSubscribers(self):
        # UNUSED CODE?
        subs = self.mychannel.subscriptions.query(format=self.context.format)
        for subscription in subs:
            self.mychannel.subscriptions.remove_subscription(subscription)
            
    def _removeSubscription(self, secret):
        """Removes subscription based on secret.
        """
        current = self.mychannel.subscriptions
        old_subscription = current.query(format=self.context.format, 
                                        secret=secret)                
        if not len(old_subscription):
            return None
        old_subscription = tuple(old_subscription)
        old_collector_data = old_subscription[0].collector_data
        for sub in old_subscription:
            current.remove_subscription(sub)
        return old_subscription

    def _addItem(self, data):
        """imports csv and returns message

        @param data: the form data        
        @return status as i18n aware unicode
        
        if a subscription with same email is found, we delete this first
        but keep selection of options.
        
        if remove-non-existing in the form was found, all existing subscriptions
        not found in the CSV file are removed.
        """

        metadata = dict(format=self.context.format,
                        date=datetime.datetime.now())

        subscriberdata = data.get('subscriberdata', None)
        remove = data.get('removenonexisting', False)
        header_row_present = data.get('header_row_present', False)
        delimiter = data.get('csv_delimiter', csv_delimiter)
        if subscriberdata is None: 
            return _(u"File was not given.")    
        subscribers, errorcandidates = parseSubscriberCSVFile(subscriberdata, 
                                self.context.composer, 
                                header_row_present=header_row_present,
                                delimiter=delimiter)
        if not type(subscribers)==type([]):
            return _(u"File has incorrect format.")
        added = 0
        new_and_updated = []
        notadded = len(errorcandidates)
        current = self.mychannel.subscriptions
        if remove:
            old = sets.Set([sub.composer_data['email'] \
                            for sub in current.values()])
        for subscriber_data in subscribers:
            secret = collective.singing.subscribe.secret(self.mychannel, 
                                                         self.context.composer, 
                                                         subscriber_data, 
                                                         self.context.request)
            try:
                old_subscription = self._removeSubscription(secret)
                item = current.add_subscription(self.mychannel, secret, 
                                                subscriber_data, [], metadata)
                new_and_updated.append(subscriber_data['email'])
                # restore section selection                
                if old_subscription is not None:
                    old_collector_data = old_subscription[0].collector_data  
                    if 'selected_collectors' in old_collector_data \
                       and old_collector_data['selected_collectors']:
                        item.collector_data = old_collector_data
                added += 1
            except Exception, e: # XXX refine which exceptions we want to catch
                # TODO; put some information about error e into the message
                errorcandidates.append(subscriber_data.get('email', 
                                                           _(u'Unknown')))
                notadded += 1 
        removed = 0
        if remove:
            for email in old.difference(sets.Set(new_and_updated)):
                for key in current.keys():
                    if key.startswith(email):
                        current.remove_subscription(current[key])
                        removed += 1                                    
                        break
        if notadded:
            msg = _(u"${numberadded} subscriptions updated successfully. "
                    u"${numberremoved} removed. "
                    u"${numbernotadded} could not be added. "
                    u"(${errorcandidates})",
                    mapping=dict(numbernotadded=str(notadded),
                                 numberremoved=str(removed),
                                 errorcandidates=', '.join(errorcandidates),
                                 numberadded=str(added))
                    )
        elif removed > 0:
            msg = _(u"${numberadded} subscriptions updated successfully, "
                    u"${numberremoved} removed!", 
                    mapping=dict(numberadded=str(added),
                                 numberremoved=str(removed))
                    )
        else:
            msg = _(u"${numberadded} subscriptions updated successfully!",
                    mapping=dict(numberadded=str(added),))        
        return msg

    @z3c.form.button.buttonAndHandler(_('Upload'), name='upload')
    def handle_add(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = form.EditForm.formErrorsMessage
            return
        try:
            self.status = self._addItem(data)
        except Exception, e:
            if DevelopmentMode:
                raise
            self.status = e            
            
    @z3c.form.button.buttonAndHandler(_('Download'), name='download')
    def handle_download(self, action):
        self.status = _(u"Subscribers exported.")
        return self.request.response.redirect(self.mychannel.absolute_url() + \
                                              '/export')            
            
class ManageUploadForm(crud.CrudForm): 
    description = _(u"Upload list of subscribers.")
    
    format = None 
    composer = None
    
    editform_factory = crud.NullForm
    addform_factory = UploadForm

    @property
    def prefix(self):
        return self.format
    

class EditComposersForm(z3c.form.form.EditForm):
    """
    """
    template = viewpagetemplatefile.ViewPageTemplateFile(
        'form-with-subforms.pt')
    subforms = []
    ignoreContext = True
    semiSuccesMessage = _(u"Only some of your changes were saved")
    
    def update(self):
        super(EditComposersForm, self).update()
        self.update_subforms()
        
    def update_subforms(self):
        self.subforms = []
        for format, item in self.context.composers.items():
             subform = component.getMultiAdapter(
                (item, self.request, self),
                z3c.form.interfaces.ISubForm)
             subform.format = format
             subform.update()
             self.subforms.append(subform)

    @z3c.form.button.buttonAndHandler(_('Save'), name='save')
    def handleSave(self, action):
        self.status = ''
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        changes = self.applyChanges(data)
        if not changes:
            self.update_subforms()
            stati = [f.status for f in self.subforms]
            if self.successMessage in stati:
                if self.formErrorsMessage not in stati:
                    self.status = self.successMessage
                else:
                    self.status = self.semiSuccesMessage
            elif self.formErrorsMessage in stati:
                self.status = self.formErrorsMessage
        if not self.status:
            if changes: 
                self.status = self.successMessage
            else:
                self.status = self.noChangesMessage


class ManageChannelView(BrowserView):
    """Manage channel view.

    Shows subscription, preview and edit options.
    """
    
    __call__ = ViewPageTemplateFile('controlpanel.pt')
    preview_form = ChannelPreviewForm
    edit_form = EditChannelForm
    composers_form = EditComposersForm
    
    @property
    def back_link(self):
        return dict(label=_(u"Up to Channels administration"),
                    url=self.context.aq_inner.aq_parent.absolute_url())

    @property
    def label(self):
        return _(u'Edit Channel ${channel}',
                 mapping={'channel':self.context.title})

    def contents(self):
        # A call to 'switch_on' is required before we can render z3c.forms.
        z2.switch_on(self,
                     request_layer=collective.singing.interfaces.IFormLayer)

        fieldsets = []

        # Add the subscriptions tab:
        forms = []
        for format, composer in self.context.composers.items():
            form = ManageSubscriptionsForm(self.context, self.request)
            form.format = format
            form.composer = composer
            forms.append(form)            

            form = ManageUploadForm(self.context, self.request)
            form.format = format
            form.composer = composer
            forms.append(form)
        fieldsets.append((_(u"Subscriptions"), '\n'.join([form() for form in forms])))

        # Add edit, composers and preview tabs:
        fieldsets.append((_(u"Edit"), self.edit_form(self.context, self.request)()))
        fieldsets.append((_(u"Composers"), self.composers_form(self.context, self.request)()))
        fieldsets.append((_(u"Preview"), self.preview_form(self.context, self.request)()))

        wrapper = """<dl class="enableFormTabbing">%s</dl>"""
        template = """
        <dt id="fieldsetlegend-%d">%s</dt>
        <dd id="fieldset-%d">
          %s
        </dd>
        """
        
        return wrapper % \
               ("\n".join((template % (id(msg), zope.i18n.translate(msg), id(msg), html)
                           for (msg, html) in fieldsets)))

        



import re
import urlparse
from BeautifulSoup import BeautifulSoup, SoupStrainer

from zope import interface
from zope import component

from Products.CMFPlone.interfaces import IPloneSiteRoot
import collective.singing.interfaces

from collective.dancing import utils

class URL(object):
    interface.implements(collective.singing.interfaces.ITransform)

    base = None
    aliases = []
    
    def __init__(self, context):
        self.context = context

    @property
    def context_url(self):
        url = self.context.absolute_url()
        if url.endswith('/'):
            return url[:-1]
        return url
        
    @property
    def site_url(self):
        site = component.getUtility(IPloneSiteRoot)
        url = utils.fix_request(site, 0).absolute_url()
        if url.endswith('/'):
            return url[:-1]
        return url
    
    def _base(self):
        if self.base is None:
            return self.site_url

        url = self.base
        if url.endswith('/'):
            return url[:-1]
        return url        

    def __call__(self, text, subscription):
        anchor_exp = re.compile('#\w+')
        root_exp = re.compile('^/')
        relative_exp = re.compile('^(?!(\w+://|mailto:|javascript:|/))')
        alias_exp = re.compile('|'.join(self.aliases), re.IGNORECASE)
        soup = BeautifulSoup(text, fromEncoding='UTF-8') # hmm
        curl = self.context.absolute_url()
        curl_parts = curl.split('/')

        for attr in ('href', 'src'):                
            for tag in soup.findAll(SoupStrainer(**{attr:root_exp})):
                if len(curl_parts) > 3 and \
                       ':' in curl_parts[2] and \
                       tag[attr].startswith('/%s/' % curl_parts[3]):
                    tag[attr] = '/'+'/'.join(tag[attr].split('/')[2:])

                # Kupu makes absolute links without the domain, which
                # include the Plone site, so let's try and strip the
                # Plone site's id out:
                site_id = component.getUtility(IPloneSiteRoot).getId()
                if tag[attr].startswith('/%s/' % site_id):
                    tag[attr] = tag[attr].replace('/%s/' % site_id, '/', 1)

                tag[attr] = '%s%s' % (self.site_url, tag[attr])

            for tag in soup.findAll(SoupStrainer(**{attr:relative_exp})):
                if tag[attr].startswith('#'):
                    tag[attr] = self.context_url + tag[attr]
                    continue
                
                parts = (self.context_url + '/'+ tag[attr]).split('/')
                while '..' in parts:
                    dots = parts.index('..')
                    del parts[dots-1:dots+1]
                tag[attr] = '/'.join(parts)

            for tag in soup.findAll(SoupStrainer(**{attr:anchor_exp})):
                prot, dom, path, params, query, frag =  urlparse.urlparse(tag[attr])

                if not prot or not dom:
                    tag[attr] = '#%s' % frag
                    continue
                
                url = '%s://%s%s' % (prot, dom, path)
                if url.endswith('/'):
                    url = url[:-1]

                # If the url points to our context and the anchor exists in our
                # text we change it to a bare anchor.
                # XXX: Maybe this should work with links to non-default views.
                if url == self.context_url:
                    for match in soup.findAll(attrs=dict(name=frag)):
                        if match.name == u'a':
                            tag[attr] = '#%s' % frag                            

            # Check for aliases
            if self.aliases:
                for tag in soup.findAll(SoupStrainer(**{attr:alias_exp})):
                    p = re.compile('^(\w+://)(%s)(/?)(.*)' %
                                   '|'.join(self.aliases),
                                   re.IGNORECASE)
                    tag[attr] = p.sub(r'%s\3\4' % self._base(), tag[attr])
            
        return str(soup)



import zope.schema.vocabulary
import z3c.form.interfaces
import Acquisition

import collective.singing.async

def get_queue():
    """Get the job queue"""
    return collective.singing.async.get_queue('collective.dancing.jobs')

def get_request_container():
    site = zope.app.component.hooks.getSite()
    return site.aq_chain[-1]

def fix_request(wrapped, skip=1):
    return aq_append(wrapped, get_request_container(), skip)

def aq_append(wrapped, item, skip=0):
    """Return wrapped with an aq chain that includes `item` at the
    end.
    
      >>> class AQ(Acquisition.Explicit):
      ...     def __init__(self, name):
      ...         self.name = name
      ...     def __repr__(self):
      ...         return '<AQ %s>' % self.name

      >>> one, two, three = AQ('one'), AQ('two'), AQ('three')
      >>> one_of_two = one.__of__(two)
      >>> one_of_two.aq_chain
      [<AQ one>, <AQ two>]
      >>> aq_append(one_of_two, three).aq_chain
      [<AQ one>, <AQ two>, <AQ three>]
      >>> aq_append(one_of_two, three, skip=1).aq_chain
      [<AQ one>, <AQ three>]
    """
    value = item
    for item in tuple(reversed(wrapped.aq_chain))[skip:]:
        value = Acquisition.aq_base(item).__of__(value)
    return value

class AttributeToDictProxy(object):
    def __init__(self, wrapped, default=z3c.form.interfaces.NOVALUE):
        super(AttributeToDictProxy, self).__setattr__('wrapped', wrapped)
        super(AttributeToDictProxy, self).__setattr__('default', default)

    def __setitem__(self, name, value):
        self.wrapped[name] = value

    __setattr__ = __setitem__

    def __getattr__(self, name):
        return self.wrapped.get(name, self.default)

class LaxVocabulary(zope.schema.vocabulary.SimpleVocabulary):
    """This vocabulary treats values the same if they're equal.
    """
    def getTerm(self, value):
        same = [t for t in self if t.value == Acquisition.aq_base(value)]
        if same:
            return same[0]
        else:
            raise LookupError(value)



import os
from setuptools import setup, find_packages
from xml.dom.minidom import parse

def readversion():
    mdfile = os.path.join(os.path.dirname(__file__), 'collective', 'dancing', 
                          'profiles', 'default', 'metadata.xml')
    metadata = parse(mdfile)
    assert metadata.documentElement.tagName == "metadata"
    return metadata.getElementsByTagName("version")[0].childNodes[0].data

def read(*pathnames):
    return open(os.path.join(os.path.dirname(__file__), *pathnames)).read()

setup(name='collective.dancing',
      version=readversion().strip(),
      description="The all-singing all-dancing newsletter product for Plone.",
      long_description='\n'.join([
          read('docs', 'README.txt'),
          read('docs', 'THANKS.txt'),
          read('docs', 'HISTORY.txt'),
          ]),
      classifiers=[
        "Framework :: Plone",
        "Framework :: Zope2",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
      keywords='zope plone notification newsletter',
      author='Daniel Nouri, Thomas Clement Mogensen and contributors',
      author_email='singing-dancing@googlegroups.com',
      url='http://plone.org/products/dancing',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['collective'],
      include_package_data=True,
      zip_safe=False,

      # If the dependency to z3c.form gives you trouble within a Zope
      # 2 environment, try the `fakezope2eggs` recipe
      install_requires=[
          'setuptools',
          'collective.singing>=0.6.12',
          'plone.z3cform>=0.5.1',
          'plone.app.z3cform>=0.4.2',
          'five.intid',
          'zc.lockfile',
          'StoneageHTML',
          'BeautifulSoup',
      ],
      
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



# --*-- coding: utf-8 --*--
import atexit
import datetime
import logging
import inspect
import md5
import re
import os
import string
import tempfile
from email.Utils import formataddr
from email.Header import Header

import stoneagehtml
from BeautifulSoup import BeautifulSoup

import persistent
from ZODB.POSException import ConflictError
from zope import interface
from zope import component
from zope import schema
import zope.annotation.interfaces
from zope.app.pagetemplate.viewpagetemplatefile import ViewPageTemplateFile
import zope.app.component.hooks
import zope.sendmail.mailer
import zope.publisher.interfaces.browser
import Products.CMFCore.interfaces
import Products.CMFPlone.interfaces
import collective.singing.interfaces
import collective.singing.mail

from collective.dancing import MessageFactory as _
from collective.dancing import transform
from collective.dancing import utils

from interfaces import IFullFormatter
from interfaces import IHTMLComposer

from plone.memoize import volatile 

logger = logging.getLogger('collective.singing')

class InvalidEmailAddress(schema.ValidationError):
    _(u"Your e-mail address is invalid")
    regex = r"[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*@(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?"

def check_email(value):
    """
      >>> def t(value):
      ...    try:
      ...        return check_email(value)
      ...    except InvalidEmailAddress:
      ...        return False
      
    Some common examples.

      >>> t('tmog@domain.tld')
      True
      >>> t('t.mog@domain.tld')
      True
      >>> t('t-mog@subdomain.domain.tld')
      True
      >>> t('tmog@sub-domain.domain.tld')
      True

    Note that we only accept real-world routeable addresses

      >>> t('tmog@localhost')
      False

    We also do not accept capitals.

      >>> t('TMOG@domain.TLD')
      False
      >>> t('Tmog@domain.tld')
      False
      >>> t('TMOG@DOMAIN.TLD')
      False
      
    This passed with the old regex.
    
      >>> t('tmog@domain.tld.')
      False

    More fails.
        
      >>> t('tmog@domain@tld')
      False
      >>> t('tmog.domain.tld')
      False
      >>> t('tmog')
      False

    No international chars plz.

      >>> t('René@la-resistance.fr')
      False

    """
    if not re.match(InvalidEmailAddress.regex, value):
        raise InvalidEmailAddress
    # Lazy: Not adding this final check to the regexp.
    return not value.endswith('.')

class PrimaryLabelTextLine(schema.TextLine):
    interface.implements(collective.singing.interfaces.ISubscriptionKey,
                         collective.singing.interfaces.ISubscriptionLabel)

    def fromUnicode(self, str):
        str = str.lower().strip()
        return super(PrimaryLabelTextLine, self).fromUnicode(str)

def _render_cachekey(method, self, vars, items):
    return (vars, items)

def template_var(varname):
    """Should return a string that is unlikely to occur in the
    rendered newsletter.  This could be made a lot more unlikely.
    Remember the result must be a valid, single string.Template
    variable name.
    """
    return '%s%s' % ('composervariable', varname)

class IHTMLComposerSchema(interface.Interface):
    email = PrimaryLabelTextLine(title=_(u"E-mail address"),
                                 constraint=check_email)

@component.adapter(collective.singing.interfaces.ISubscription)
@interface.implementer(IHTMLComposerSchema)
def composerdata_from_subscription(subscription):
    return utils.AttributeToDictProxy(subscription.composer_data)

class HTMLComposer(persistent.Persistent):
    """
      >>> from zope.interface.verify import verifyClass
      >>> from collective.dancing.interfaces import IHTMLComposer
      >>> verifyClass(IHTMLComposer, HTMLComposer)
      True
    """

    interface.implements(IHTMLComposer)

    title = _(u'HTML E-Mail')
    schema = IHTMLComposerSchema

    encoding = 'utf-8'
    stylesheet = u""
    from_name = u""
    from_address = u""
    replyto_address = u""
    subject = u"${site_title}: ${channel_title}"
    header_text = u"<h1>${subject}</h1>"
    footer_text = u""
    
    template = ViewPageTemplateFile('browser/composer-html.pt')
    confirm_template = ViewPageTemplateFile('browser/composer-html-confirm.pt')
    forgot_template = ViewPageTemplateFile('browser/composer-html-forgot.pt')

    @staticmethod
    def secret(data):
        salt = component.getUtility(collective.singing.interfaces.ISalt)
        return md5.new("%s%s" % (data['email'], salt)).hexdigest()
    
    context = None
    @property
    def request(self):
        site = zope.app.component.hooks.getSite()
        return site.REQUEST

    @property
    def _from_address(self):
        properties = component.getUtility(
            Products.CMFCore.interfaces.IPropertiesTool)
        charset = properties.site_properties.getProperty('default_charset', 'utf-8')

        name = self.from_name or properties.email_from_name
        mail = self.from_address or properties.email_from_address

        if not isinstance(name, unicode):
            name = name.decode(charset)
        if not isinstance(mail, unicode):
            # mail has to be be ASCII!!
            mail = mail.decode(charset).encode('us-ascii', 'replace')
            #TODO : assert that mail is now valid. (could have '?' from repl.)

        return formataddr((str(Header(name, charset)), mail))

    @property
    def language(self):
        return self.request.get('LANGUAGE')        

    def _vars(self, subscription, items=()):
        """Provide variables for the template.

        Override this or '_more_vars' in your custom HTMLComposer to
        pass different variables to the templates.
        """
        vars = {}
        site = component.getUtility(Products.CMFPlone.interfaces.IPloneSiteRoot)
        site = utils.fix_request(site, 0)
        fix_urls = lambda t: transform.URL(site).__call__(t, subscription)
        
        vars['channel'] = subscription.channel
        vars['site_url'] = site.absolute_url()
        vars['site_title'] = unicode(site.Title(), 'UTF-8')
        vars['channel_title'] = subscription.channel.title
        vars['subject'] = self.subject
        # Why would header_text or footer_text ever be None?
        vars['header_text'] = fix_urls(self.header_text or u"")
        vars['footer_text'] = fix_urls(self.footer_text or u"")
        vars['stylesheet'] = self.stylesheet
        vars['from_addr'] = self._from_address
        headers = vars['more_headers'] = {}
        if self.replyto_address:
            headers['Reply-To'] = self.replyto_address
        
        # This is so brittle, it hurts my eyes.  Someone convince me
        # that this needs to become another component:
        for index, item in enumerate(items):
            formatted, original = item
            title = getattr(original, 'Title', lambda: formatted)()
            vars['item%s_title' % index] = title

        vars.update(self._more_vars(subscription, items))

        def subs(name):
            vars[name] = string.Template(vars[name]).safe_substitute(vars)
        for name in 'subject', 'header_text', 'footer_text':
            subs(name)

        # It'd be nice if we could use an adapter here to override
        # variables.  We'd probably want to pass 'items' along to that
        # adapter.

        return vars    

    def _more_vars(self, subscription, items):
        """Less generic variables.
        """
        vars = {}
        channel = subscription.channel
        site = component.getUtility(Products.CMFPlone.interfaces.IPloneSiteRoot)
        site = utils.fix_request(site, 0)
        secret_var = '$%s' % template_var('secret')
        vars['confirm_url'] = (
            '%s/confirm-subscription.html?secret=%s' %
            (site.portal_newsletters.absolute_url(), secret_var))
        vars['unsubscribe_url'] = (
            '%s/unsubscribe.html?secret=%s' %
            (channel.absolute_url(), secret_var))
        vars['my_subscriptions_url'] = (
            '%s/../../my-subscriptions.html?secret=%s' %
            (channel.absolute_url(), secret_var))
        vars['to_addr'] = '$%s' % template_var('to_addr')
        return vars

    def _subscription_vars(self, subscription):
        """Variables that are expected to be unique
        to every subscription.
        """
        vars = {}
        vars[template_var('secret')] = self.secret(subscription.composer_data)
        vars[template_var('to_addr')] = subscription.composer_data['email']
        for k, v in subscription.composer_data.items():
            vars[template_var(k)] = v
        return vars

    @volatile.cache(_render_cachekey)
    def _render(self, vars, items):
        html = self.template(
            contents=[i[0] for i in items],
            items=[dict(formatted=i[0], original=i[1]) for i in items],
            **vars)
        return stoneagehtml.compactify(html).decode('utf-8')

    def render(self, subscription, items=(), override_vars=None):
        vars = self._vars(subscription, items)
        subscription_vars = self._subscription_vars(subscription)

        if override_vars is None:
            override_vars = {}
        vars.update(override_vars)

        html = self._render(vars, items)
        html = string.Template(html).safe_substitute(subscription_vars)

        message = collective.singing.mail.create_html_mail(
            vars['subject'],
            html,
            from_addr=vars['from_addr'],
            to_addr=subscription_vars[template_var('to_addr')],
            headers=vars.get('more_headers'),
            encoding=self.encoding)

        return collective.singing.message.Message(
            message, subscription)

    def render_confirmation(self, subscription):
        vars = self._vars(subscription)
        subscription_vars = self._subscription_vars(subscription)
        
        if 'confirmation_subject' not in vars:
            vars['confirmation_subject'] = zope.i18n.translate(
                _(u"Confirm your subscription with ${channel-title}",
                  mapping={'channel-title': subscription.channel.title}),
                target_language=self.language)

        html = self.confirm_template(**vars)
        html = stoneagehtml.compactify(html).decode('utf-8')
        html = string.Template(html).safe_substitute(subscription_vars)
        
        message = collective.singing.mail.create_html_mail(
            vars['confirmation_subject'],
            html,
            from_addr=vars['from_addr'],
            to_addr=subscription_vars[template_var('to_addr')],
            headers=vars.get('more_headers'),
            encoding=self.encoding)

        # status=None prevents message from ending up in any queue
        return collective.singing.message.Message(
            message, subscription, status=None)

    def render_forgot_secret(self, subscription):
        vars = self._vars(subscription)
        subscription_vars = self._subscription_vars(subscription)
        
        if 'forgot_secret_subject' not in vars:
            vars['forgot_secret_subject'] = zope.i18n.translate(
                _(u"Change your subscriptions with ${site-title}",
                  mapping={'site-title': vars['site_title']}),
                target_language=self.language)

        html = self.forgot_template(**vars)
        html = stoneagehtml.compactify(html).decode('utf-8')
        html = string.Template(html).safe_substitute(subscription_vars)

        message = collective.singing.mail.create_html_mail(
            vars['forgot_secret_subject'],
            html,
            from_addr=vars['from_addr'],
            to_addr=subscription_vars[template_var('to_addr')],
            headers=vars.get('more_headers'),
            encoding=self.encoding)

        # status=None prevents message from ending up in any queue
        return collective.singing.message.Message(
            message, subscription, status=None)

plone_html_strip_not_likey = [
    {'id': 'review-history'},
    {'class':'documentActions'},
    {'class':'portalMessage'},
    {'id':'plone-document-byline'},
    {'id':'portlets-below'},
    {'id':'portlets-above'},
    {'class': 'newsletterExclude'},
    ]
def plone_html_strip(html, not_likey=plone_html_strip_not_likey):
    r"""Tries to strip the relevant parts from a Plone HTML page.

    Looks for ``<div id="content">`` and ``<div id="region-content">`` as
    a fallback.

      >>> html = (
      ...     '<html><body><div id="content"><h1 class="documentFirstHeading">'
      ...     'Hi, it\'s me!</h1><p>Wannabe the son of Frankenstein</p>'
      ...     '</div></body></html>')
      >>> plone_html_strip(html)
      u'<h1 class="documentFirstHeading">Hi, it\'s me!</h1><p>Wannabe the son of Frankenstein</p>'
      >>> plone_html_strip('<div id="region-content">Hello, World!</div>')
      u'Hello, World!'

    Will also strip away any ``<div id="review-history">``:

      >>> html = (
      ...     '<div id="region-content">'
      ...     '<div id="review-history">Yesterday</div>Tomorrow</div>')
      >>> plone_html_strip(html)
      u'Tomorrow'

    """
    if not isinstance(html, unicode):
        html = unicode(html, 'UTF-8')

    soup = BeautifulSoup(html)
    content = soup.find('div', attrs={'id': 'content'})
    if content is None:
        content = soup.find('div', attrs=dict({'id': 'region-content'}))

    for attrs in not_likey:
        for item in content.findAll(attrs=attrs):
            item.extract() # meaning: item.bye()
    return content.renderContents(encoding=None) # meaning: as unicode

class CMFDublinCoreHTMLFormatter(object):
    """Render a brief representation of an IBaseContent for HTML.
    """
    interface.implements(collective.singing.interfaces.IFormatItem)
    component.adapts(Products.CMFCore.interfaces.IMinimalDublinCore,
                     zope.publisher.interfaces.browser.IBrowserRequest)

    template = """\
    <div>
      <h2><a href="%(url)s">%(title)s</a></h2>
      <p>%(description)s</p>
    </div>
    """
    
    def __init__(self, item, request):
        self.item = item
        self.request = request

    def __call__(self):
        i = self.item
        return self.template % dict(
            url=i.absolute_url(), title=i.Title(), description=i.Description())

class PloneCallHTMLFormatter(object):
    """Assumes that item is callable and returns an HTML
    representation.

    If what ``item()`` returns looks like a rendered Plone page, this
    formatter will try and strip away all irrelevant parts.
    """
    
    interface.implements(collective.singing.interfaces.IFormatItem)

    def __init__(self, item, request):
        self.item = item

    def __call__(self):
        try:
            html = self.item()
        except ConflictError:
            raise
        except:
            # Simple calling does not work if layout
            # is a view, we can try a little harder...
            html = self.item.unrestrictedTraverse(self.item.getLayout())()
        if 'kss' in html:
            return plone_html_strip(html)
        else:
            return html

class FullFormatWrapper(object):
    """Wraps an item for use with a full formatter."""
    
    def __init__(self, item):
        self.item = item

    def __getattr__(self, key):
        return getattr(self.item, key)

class HTMLFormatItemFully(object):
    interface.implements(collective.singing.interfaces.IFormatItem)
    
    def __init__(self, wrapper, request):
        self.item = wrapper.item
        self.request = request
                 
    def __call__(self):
        view = component.getMultiAdapter(
            (self.item, self.request), IFullFormatter, name='html')

        return view()

class SMTPMailer(zope.sendmail.mailer.SMTPMailer):
    """A mailer for use with zope.sendmail that fetches settings from
    the Plone site's configuration.
    """
    def _fetch_settings(self):
        root = component.getUtility(Products.CMFPlone.interfaces.IPloneSiteRoot)
        m = root.MailHost
        return dict(hostname=m.smtp_host or 'localhost',
                    port=m.smtp_port,
                    username=m.smtp_userid or m.smtp_uid or None,
                    password=m.smtp_pass or m.smtp_pwd or None,)

    def send(self, fromaddr, toaddrs, message):
        self.__dict__.update(self._fetch_settings())
        return super(SMTPMailer, self).send(fromaddr, toaddrs, message)

class StubSMTPMailer(zope.sendmail.mailer.SMTPMailer):
    """An ISMPTMailer that'll only log what it would do.
    """
    logfile = None
    sent = 0
    recipients = set()

    def __init__(self, *args, **kwargs):
        super(StubSMTPMailer, self).__init__(*args, **kwargs)
        path = os.path.join(tempfile.gettempdir(),
                            'collective.dancing.StubSMTPMailer.log')
        StubSMTPMailer.logfile = logfile = open(path, 'a')
        
        logger.info("%r logging to %s" % (self, path))
        self.log("StubSMTPMailer starting to log.")

    def send(self, fromaddr, toaddrs, message):
        self.log("StubSMTPMailer.send From: %s, To: %s\n" % (fromaddr, toaddrs))
        StubSMTPMailer.sent += 1
        self.recipients.add(toaddrs)

    @staticmethod
    def log(msg):
        StubSMTPMailer.logfile.write(
            "%s: %s\n" % (datetime.datetime.now(), msg))
        StubSMTPMailer.logfile.flush()

@atexit.register
def at_exit():
    if StubSMTPMailer.logfile is None:
        return
    StubSMTPMailer.log(
        "StubSMTPMailer shutting down, sent %s messages to %s recipients.\n" %
        (StubSMTPMailer.sent, len(StubSMTPMailer.recipients)))
    StubSMTPMailer.logfile.close()



import persistent.dict
from Acquisition import aq_parent, aq_inner
from zope import component
from zope import interface
from zope import schema
from plone.portlets.interfaces import IPortletDataProvider
from plone.app.portlets.portlets import base
from zope.app.pagetemplate import viewpagetemplatefile
from zope.formlib import form

from Products.CMFCore.utils import getToolByName
from urllib import urlencode
from Products.Five import BrowserView
import z3c.form
from z3c.form import subform
from plone.z3cform import z2

from plone.memoize.instance import memoize
from plone.memoize import ram
from plone.memoize.compress import xhtml_compress

from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile

import zope.app.container
import collective.singing
from collective.singing.channel import channel_lookup
from collective.dancing import MessageFactory as _
from collective.dancing.browser.subscribe import SubscriptionAddForm
from collective.singing.interfaces import ICollectorSchema


test_vocab = schema.vocabulary.SimpleVocabulary.fromValues(range(5))
 
class IChannelSubscribePortlet(IPortletDataProvider):
    """A portlet which renders the results of a collection object.
    """

    header = schema.TextLine(title=_(u"Portlet header"),
                             description=_(u"Title of the rendered portlet"),
                             required=True)
    channel = schema.Choice(title=_(u"The channel to enable subscriptions to."),
                            description=_(u"Find the channel you want to enable direct subscription to"),
                            required=False,
                            vocabulary='collective.singing.vocabularies.ChannelVocabulary'
                            )
    description = schema.TextLine(title=_(u"Portlet description"),
                           description=_(u"Description of the rendered portlet"),
                           required=True)
    subscribe_directly = schema.Bool(title=_(u"Subscribe directly from portlet"),
                            description=_(u"Click here to select collector options to be automatically enabled, when subscribing from this portlet."),
                            required=True,
                            default=True)
    footer_text = schema.TextLine(title=_(u"Footer text"),
                             description=_(u"Text in footer - if omitted the channel title is used"),
                             required=False)

    show_footer = schema.Bool(title=_(u"Show footer"),
                            description=_(u"Click here to show the portlet footer."),
                            required=True,
                            default=True)

class Assignment(base.Assignment):
    """
    Portlet assignment.    
    This is what is actually managed through the portlets UI and associated
    with columns.
    """
    interface.implements(IChannelSubscribePortlet, ICollectorSchema)
    header = u""
    descriptions = False
    channel=None
    subscribe_directly = True

    def __init__(self,
                 header=u"",
                 description=u"",
                 channel=None,
                 subscribe_directly=True,
                 footer_text="",
                 show_footer=True):
        self.header = header
        self.description = description
        self._channel = channel
        self.subscribe_directly = subscribe_directly
        self.footer_text = footer_text
        self.show_footer = show_footer

    @apply
    def channel():
        def get(self):
            # BBB: Versions prior to r67243 used to have an attribute
            # called 'channel'; then it's become a property
            channel = self.__dict__.get('channel')
            if channel is None:
                channel = self._channel
            if channel is not None:
                for c in self.all_channels:
                    if c.name == channel.name:
                        return c
            return None
        def set(self, value):
            self._channel = value
        return property(get, set)


    @property
    def title(self):
        """This property is used to give the title of the portlet in the
        "manage portlets" screen. Here, we use the title that the user gave.
        """
        return self.header

    @property
    def all_channels(self):
        return channel_lookup(only_subscribeable=True)
        
class ValuesMixin(object):
    """Mix-in class that allows convenient access to data stored on
    the assignment."""

    channel_id = None
    assignment = None

    @apply
    def stored_values():
        def get(self):
            d = getattr(self.assignment, '_stored_values',
                        persistent.dict.PersistentDict())
            return d.setdefault(
                self.channel_id, persistent.dict.PersistentDict())
        def set(self, value):
            d = getattr(self.assignment, '_stored_values',
                        persistent.dict.PersistentDict())
            d[self.channel_id] = value
            self.assignment._stored_values = d
        return property(get, set)
    
class PortletSubscriptionAddForm(ValuesMixin, SubscriptionAddForm):
    """ """
    template = viewpagetemplatefile.ViewPageTemplateFile('htmlstatusform.pt')

    assignment = None

    @property
    def status_already_subscribed(self):
        return _(u'You are already subscribed to this newsletter. Click here to <a href="${url}">edit your subscriptions</a>.',
                 mapping={'url':
                          '%s/sendsecret.html' % self.newslettertool.absolute_url()})

    @property
    def newslettertool(self):
        return getToolByName(self.context, 'portal_newsletters')
    
    def update(self):
        super(PortletSubscriptionAddForm, self).update()
        self.channel_id = self.context.id

        if self.context.collector is not None:
            stored_values = self.stored_values
            collector_schema = self.context.collector.schema

            for name in collector_schema.names():
                field = collector_schema.get(name)
                widget = self.widgets['collector.' + name]
                value = stored_value = stored_values.get(name)
#                 if stored_value is not None:
#                     subfield = field
#                     vocabulary = None
#                     while (not hasattr(subfield, 'vocabulary')) and \
#                               (hasattr(subfield, 'value_type')):
#                         subfield = subfield.value_type
#                     if hasattr(subfield, 'vocabulary'):
#                         if 
#                         value = set([v for v in stored_value
#                                      if v in subfield.vocabulary])
                if value:
                    converter = z3c.form.interfaces.IDataConverter(widget)
                    widget.value = converter.toWidgetValue(stored_value)
#                else:
#                    widget.value = value
                widget.update()

    @property
    def fields(self):
        fields = z3c.form.field.Fields(
            self.context.composers[self.format].schema,
            prefix='composer.')
        if self.context.collector is not None and self.assignment.subscribe_directly:
            fields += z3c.form.field.Fields(self.context.collector.schema,
                                            prefix='collector.',
                                            mode=z3c.form.interfaces.HIDDEN_MODE)
        return fields

class PortletSubscribeLinkForm(z3c.form.form.Form):
    """ """
    template = viewpagetemplatefile.ViewPageTemplateFile('titlelessform.pt')
    ignoreContext = True
    formErrorsMessage = _('There were some errors.')
    
    def __init__(self, context, request):
        super(PortletSubscribeLinkForm, self).__init__(context, request)

    @property
    def fields(self):
        return z3c.form.field.Fields(self.context.composers[self.format].schema)

    @z3c.form.button.buttonAndHandler(_('Proceed'), name='preceed')
    def handleAdd(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        params = urlencode(dict([('composer.widgets.%s'%key, value)
                                 for key, value in data.items()]))
        subscribe_url = '%s/subscribe.html?%s' % (self.context.absolute_url(),
                                                  params)
        self.request.response.redirect(subscribe_url)
        return 
    
class Renderer(base.Renderer):
    """Portlet renderer.
    
    This is registered in configure.zcml. The referenced page template is
    rendered, and the implicit variable 'view' will refer to an instance
    of this class. Other methods can be added and referenced in the template.
    """
    _template = ViewPageTemplateFile('channelsubscribe.pt')
    form_template = viewpagetemplatefile.ViewPageTemplateFile('titlelessform.pt')

    def __init__(self, *args):
        base.Renderer.__init__(self, *args)
        self.setup_form()

    def setup_form(self):
        z2.switch_on(self)
        if self.channel is not None:
            if self.data.subscribe_directly:
                self.form = PortletSubscriptionAddForm(self.channel, self.request)
                self.form.assignment = self.data
            else:
                self.form = PortletSubscribeLinkForm(self.channel, self.request)
            self.form.format = 'html'
            self.form.update()
        
    render = _template

    @property
    def available(self):
        return bool(self.data.channel)

    @property
    def channel(self):
        return self.data.channel
        channels = self.data.all_channels
        if channels and self.data.channel:
            for channel in channels:
                if channel.name == self.data.channel.name:
                    return channel
        return None                 

    def channel_link(self):

        link = {'url':'%s/subscribe.html'%self.channel.absolute_url(),
                'title':self.getFooterText()}
        return link

        
    def getFooterText(self):
        if bool(self.data.footer_text):
            return self.data.footer_text
        return self.channel.Title()
 
            
def prefix(self):
    return str(self.__class__.__name__ + '-'.join(self.context.getPhysicalPath()))


class EditCollectorOptionsForm(ValuesMixin, z3c.form.subform.EditSubForm):
    """Edit a single collectors options."""
    template = viewpagetemplatefile.ViewPageTemplateFile(
        '../subform.pt')

    css_class = 'subForm subForm-level-1'
    ignoreContext = True

    @property
    def heading(self):
        return _(u"${channel} options", mapping={'channel':self.context.Title()})

    prefix = property(prefix)

    @property
    def fields(self):
        return z3c.form.field.Fields(self.context.collector.schema)


    @property
    def channel_id(self):
        return self.context.id

    @property
    def assignment(self):
        return self.parentForm.context

    @z3c.form.button.handler(z3c.form.form.EditForm.buttons['apply']) 
    def handleApply(self, action): 
        data, errors = self.extractData()
        if errors: 
            self.status = self.formErrorsMessage 
            return

        stored_values = self.stored_values
        changed = False 

        for name, widget_value in data.items():
            if stored_values.get(name) == widget_value:
                continue
            else:
                stored_values[name] = widget_value
            changed = True 

        if changed: 
            self.stored_values = stored_values
            self.status = self.successMessage 
        else: 
            self.status = self.noChangesMessage

    def update(self):
        super(EditCollectorOptionsForm, self).update()

        # We add some logic here to set widget values from stored
        # values if they weren't provided in the request.  Note that
        # we have ``ignoreContext = True``.

        stored_values = self.stored_values
        for name in self.context.collector.schema.names(): 
            field = self.context.collector.schema.get(name) 
            widget = self.widgets[name]
            stored_value = stored_values.get(name)
            widget_value = widget.extract()

            if (widget_value is z3c.form.interfaces.NOVALUE and
                stored_value is not None):

                # discard values that are not in this collectors vocabulary
                # The collector may be changed or entirely different since
                # last save.
                subfield = field
                vocabulary = None
                while (not hasattr(subfield, 'vocabulary')) and \
                          (hasattr(subfield, 'value_type')):
                    subfield = subfield.value_type
                if hasattr(subfield, 'vocabulary'):                
                    value = set([v for v in stored_value
                                 if v in subfield.vocabulary])
                    
                if len(value):
                    converter = z3c.form.interfaces.IDataConverter(widget)
                    widget.value = converter.toWidgetValue(value)
                else:
                    widget.value = value
                widget.update()
    
class ChannelSubscribePortletEditForm(z3c.form.form.EditForm):
    """  """
    template = viewpagetemplatefile.ViewPageTemplateFile('../form-with-subforms.pt')
    fields = z3c.form.field.Fields(IChannelSubscribePortlet)

    css_class = 'editForm portletEditForm'
    heading = _(u"Edit Channel Subscribe Portlet")

    def update(self):
        super(ChannelSubscribePortletEditForm, self).update()
        self.subforms = []
        for channel in self.context.all_channels:
            if channel.collector is not None:
                option_form = EditCollectorOptionsForm(channel,
                                                       self.request,
                                                       self)
                option_form.update()
                self.subforms.append(option_form)
        

class ChannelSubscribePortletView(BrowserView):
    __call__ = ViewPageTemplateFile('z3c.plone.portlet.pt')

    def referer(self):
        return self.request.get('referer', '')

    # eventually replace this with a referer and redirect like regular
    # plone portlets.
    # NB: this would require combining status messages from subforms.
    def back_link(self):
        url = self.request.form.get('referer')
        if not url:
            addview = aq_parent(aq_inner(self.context))
            context = aq_parent(aq_inner(addview))
            url = str(component.getMultiAdapter((context, self.request),
                        name=u"absolute_url")) + '/@@manage-portlets'
        return dict(url=url,
                    label=_(u"Back to portlets"))


class ChannelSubscribePortletEditView(ChannelSubscribePortletView):

    label = _(u"Edit Channel Subscribe Portlet")
    description = _(u"This portlet allows a visitor to subscribe to a specific newsletter channel.")

    def contents(self):
        z2.switch_on(self)
        return ChannelSubscribePortletEditForm(self.context, self.request)()


class EditCollectorOptionsAddForm(z3c.form.form.Form):
    """Edit a single collectors options.
    """
    template = viewpagetemplatefile.ViewPageTemplateFile(
        '../subform.pt')

    css_class = 'subForm subForm-level-1'
    ignoreContext = True    
    prefix = property(prefix)
    
    def __init__(self, context, request, channel, parentForm):
        super(EditCollectorOptionsAddForm, self).__init__(context, request)
        self.context = context
        self.request = request
        self.channel = channel
        self.parentForm = self.__parent__ = parentForm
        self.heading = 'Options for %s'%channel.Title()
    
    @property
    def fields(self):
        return z3c.form.field.Fields(self.channel.collector.schema)

    @property
    def label(self):
        return _(u"${channel} options", mapping={'channel':self.channel.Title()})

    @property
    def selected_channel(self):
        return self.context == self.parentForm.context.channel

    
class ChannelSubscribePortletAddForm(z3c.form.form.AddForm):
    """ """
    template = viewpagetemplatefile.ViewPageTemplateFile('../form-with-subforms.pt')
    fields = z3c.form.field.Fields(IChannelSubscribePortlet)

    css_class = 'addForm portletAddForm'
    heading = _(u"Add Channel Subscribe Portlet")

    subforms = []
    
    def create(self, data):
        return Assignment(**data)

    def add(self, object):
        self.context.add(object)

    def nextURL(self):
        # XXX: this should be prettier/more stable
        subscribe_directly = self.request.get(
            'form.widgets.subscribe_directly', '') == [u'true']
        if subscribe_directly:
            return '../%s/edit' % (self.context.items()[-1][0])
        else:
            return '../../@@manage-portlets'
        
     
class ChannelSubscribePortletAddView(ChannelSubscribePortletView):

    label = _(u"Add Channel Subscribe Portlet")
    description = _(u"This portlet allows a visitor to subscribe to a specific newsletter channel.")

    def contents(self):
        z2.switch_on(self)
        return ChannelSubscribePortletAddForm(self.context, self.request)()




from zope import interface
from zope import schema

import collective.singing.interfaces

from collective.dancing import MessageFactory as _

class IFullFormatter(interface.Interface):
    """Format the item for use as main newsletter content.

    This is used when newsletters are created from an existing content
    object on the site.

    The formatter should only return the actual content markup, not a
    complete HTML document.
    """

class IHTMLComposer(collective.singing.interfaces.IComposer,
                    collective.singing.interfaces.IComposerBasedSecret):
    """An HTML composer.
    """

    from_name = schema.TextLine(
        title=_(u"From name"),
        description=_(u"The sender name that will appear in messages sent from this composer."),
        required=False,
        )
    from_address = schema.TextLine(
        title=_(u"From address"),
        description=_(u"The from addess that will appear in messages sent from this composer."),
        required=False,
        )
    subject = schema.TextLine(
        title=_(u"Message subject"),
        description=_(u"You may use a number of variables here, like "
                      "${site-title}."),
        required=False,
        )
    replyto_address = schema.TextLine(
        title=_(u"Reply-to address"),
        description=_(u"The reply-to addess that will appear in messages sent from this composer."),
        required=False,
        )
    header_text = schema.Text(
        title=_(u"Header text"),
        description=_(u"Text to appear at the beginning of every message. "
                      "You may use a number of variables here including "
                      "${subject}."),
        required=False,
        )
    footer_text = schema.Text(
        title=_(u"Footer text"),
        description=_(u"Text to appear at the end of every message"),
        required=False,
        )
    stylesheet = schema.Text(
        title=_(u"CSS Stylesheet"),
        description=_(u"The stylesheet will be applied to the document."),
        required=False,
        )



from Acquisition import aq_base

import collective.singing.subscribe

class Subscription(collective.singing.subscribe.SimpleSubscription):
    _channel = None
    @apply
    def channel():
        def get(self):
            if self._channel is not None:
                all_channels = list(collective.singing.channel.channel_lookup())
                # We want to get the same channel from the
                # ``channel_lookup`` as that has the correct wrapping:
                for channel in all_channels:
                    if aq_base(channel) is aq_base(self._channel):
                        return channel

                # Our channel doesn't exist!  Look for a channel with
                # the same name:
                for channel in all_channels:
                    if channel.name == self._channel.name:
                        self._channel = channel
                        return channel

            raise AttributeError(
                "%r is subscribed to unknown channel %r" % self._channel)
                    
        def set(self, channel):
            self._channel = channel
        return property(get, set)

class Subscriptions(collective.singing.subscribe.Subscriptions):
    subscription_factory = Subscription



import random
import string
from UserString import UserString

import persistent.dict
from zope import component
from zope import interface
from zope import schema
import zope.app.container.interfaces
import zope.app.component.hooks
import AccessControl
from Acquisition import aq_base
import OFS.event
import OFS.Folder
import OFS.SimpleItem
import Products.CMFPlone.interfaces
import collective.singing.interfaces
import collective.singing.message
import collective.singing.channel
import collective.singing.subscribe
import collective.dancing.collector
import collective.dancing.composer
import collective.dancing.subscribe
import collective.dancing.utils
from collective.dancing import MessageFactory as _

def portal_newsletters():
    """Return channels created with the newsletter tool."""

    root = component.queryUtility(Products.CMFPlone.interfaces.IPloneSiteRoot)
    if root is None:
        return []
    root = collective.dancing.utils.fix_request(root, 0)
    if 'portal_newsletters' in root.objectIds():
        channels = root['portal_newsletters']['channels'].objectValues()
        security = AccessControl.getSecurityManager()
        return [c for c in channels if
                security.checkPermission('View', c) and
                collective.singing.interfaces.IChannel.providedBy(c)]
    else:
        return []

interface.directlyProvides(portal_newsletters,
                           collective.singing.interfaces.IChannelLookup)

class Salt(UserString):
    """
      >>> len(Salt())
      50
    """
    interface.implements(collective.singing.interfaces.ISalt)

    def __init__(self, data=''):
        if data:
            UserString.__init__(self, data)
            return
        salt = ''.join([
            random.choice(string.ascii_letters + string.digits)
            for i in range(50)])
        UserString.__init__(self, salt)

class INewslettersSettings(interface.Interface):
    use_single_form_subscriptions_page = schema.Bool(
        title=_(u"Use single form subscriptions page"),
        description=_(u"Use single form subscriptions page when possible."),
        default=False,
        required=False)

class IPortalNewsletters(INewslettersSettings):
    pass

class PortalNewsletters(OFS.Folder.Folder):
    interface.implements(IPortalNewsletters)

    use_single_form_subscriptions_page = False

    def Title(self):
        return u"Newsletters"

@component.adapter(IPortalNewsletters,
                   zope.app.container.interfaces.IObjectAddedEvent)
def tool_added(tool, event):
    # Add children
    factories = dict(channels=ChannelContainer,
                     collectors=collective.dancing.collector.CollectorContainer)
    existing = tool.objectIds()
    for name, factory in factories.items():
        if name not in existing:
            tool[name] = factory(name)

    # Create and register salt
    salt = getattr(aq_base(tool), 'salt', Salt())
    tool.salt = salt
    sm = zope.component.getSiteManager(tool)
    sm.registerUtility(salt, collective.singing.interfaces.ISalt)

class IChannelContainer(interface.Interface):
    pass

class ChannelContainer(OFS.Folder.Folder):
    interface.implements(IChannelContainer)
    """
      >>> container = ChannelContainer('xs')
      >>> container['your-channel'] = Channel('wq')
      >>> container.objectIds()
      ['your-channel']
    """
    def Title(self):
        return u"Channels"

@component.adapter(IChannelContainer,
                   zope.app.container.interfaces.IObjectAddedEvent)
def channels_added(container, event):
    if 'default-channel' not in container.objectIds():
        default_channel = Channel('default-channel', title=_(u"Newsletter"))
        container['default-channel'] = default_channel

class Channel(OFS.SimpleItem.SimpleItem):
    """
      >>> channel = Channel('xs')
      >>> from zope.interface.verify import verifyObject
      >>> verifyObject(collective.singing.interfaces.IChannel, channel)
      True
      >>> channel.name
      'xs'
    """
    interface.implements(collective.singing.interfaces.IChannel)

    type_name = _("Standard Channel")

    subscribeable = True
    
    sendable = True
    
    def __init__(self, name, title=None, composers=None,
                 collector=None, scheduler=None, description=u""):
        self.name = name
        if title is None:
            title = name
        self.title = title
        self.description = description
        self.subscriptions = collective.dancing.subscribe.Subscriptions()
        if composers is None:
            composers = persistent.dict.PersistentDict()
            composers['html'] = collective.dancing.composer.HTMLComposer()
        self.composers = composers
        self.collector = collector
        self.scheduler = scheduler
        self.queue = collective.singing.message.MessageQueues()
        super(Channel, self).__init__(name)

    @property
    def id(self):
        return self.name

    def Title(self):
        return self.title

@component.adapter(Channel,
                   zope.app.container.interfaces.IObjectAddedEvent)
def channel_added(channel, event):
    # We'll take extra care that when we're imported through the ZMI,
    # we update things to keep everything up to date:
    subscriptions = collective.singing.subscribe.subscriptions_data(channel)
    subscriptions._catalog.clear()
    
    for subscription in subscriptions.values():
        # Let's make sure that ``subscription.channel`` refers to the
        # channel:
        if aq_base(subscription.channel) is not aq_base(channel):
            subscription.channel = channel

        # The secret may have changed:
        composer = channel.composers[subscription.metadata['format']]
        secret = composer.secret(subscription.composer_data)
        if subscription.secret != secret:
            subscription.secret = secret

        # This will finally catalog the subscription:
        collective.singing.subscribe.subscription_added(subscription, None)

@component.adapter(collective.singing.interfaces.ICollector,
                   zope.app.container.interfaces.IObjectRemovedEvent)
def collector_removed(collector, event):
    for channel in collective.singing.channel.channel_lookup():
        if isinstance(channel, Channel):
            if aq_base(channel.collector) is aq_base(collector):
                channel.collector = None

# This lists of factories is mutable: You can add to it:
channels = [Channel,]



from zope import interface
from zope import component
from zope import schema

from Products.Five.browser import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile 
from Products.statusmessages.interfaces import IStatusMessage
from zope.app.publisher.browser import getDefaultViewName

from collective.dancing.composer import FullFormatWrapper
from collective.singing.channel import lookup
from collective.singing.interfaces import ISubscription
from collective.singing.interfaces import IChannel
from collective.singing.interfaces import IMessageAssemble

from collective.dancing import MessageFactory as _

import transaction

class PreviewSubscription(object):
    interface.implements(ISubscription)

    secret = u""
    format = 'html'
    
    def __init__(self, channel):
        self.channel = channel

        self.collector_data = {}
        self.metadata = dict(format=self.format)
        
        composer = self.channel.composers[self.format]
        
        # set default composer data
        self.composer_data = dict(
            (name, field.default) \
            for name, field in schema.getFields(composer.schema).items())
        
class PreviewNewsletterView(BrowserView):
    template = ViewPageTemplateFile("preview.pt")
    
    def __call__(self, name=None, include_collector_items=False, override_vars=None):

        if override_vars is None:
            override_vars = '{}'

        if IChannel.providedBy(self.context):
            channel = self.context
            items = ()
        else:
            assert name is not None
            channel = lookup(name)
            items = (FullFormatWrapper(self.context),)
            
        sub = PreviewSubscription(channel)

        # We don't want to commit a transaction just for the preview;
        # this would happen if we were to simply add and remove from
        # the queue.
        sp = transaction.savepoint()

        message = IMessageAssemble(channel).render_message(
            self.request,
            sub,
            items,
            bool(include_collector_items),
            eval(override_vars))

        if message is None:
            IStatusMessage(self.request).addStatusMessage(
                _(u"No items found."))

            return self.request.response.redirect(self.context.absolute_url())

        # pull message out of hat
        channel.queue[message.status].pull(-1)

        # rollback savepoint
        sp.rollback()

        # walk message, decoding HTML payload
        for part in message.payload.walk():
            if part.get_content_type() == 'text/html':
                html = part.get_payload(decode=True)
                break
        else:
            raise ValueError("Message does not contain a 'text/html' part.")
            
        return self.template(content=html, title=channel.title)



import zope.publisher
from zope import schema
import zope.schema.vocabulary
from zope import component
from zope import interface
from zope.app.pagetemplate import viewpagetemplatefile
from z3c.form import field
from z3c.form import form, subform
import z3c.form.interfaces
import z3c.form.browser.select
import z3c.formwidget.query.widget
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
import Products.CMFPlone.utils
from collective.singing.interfaces import ICollector
from plone.z3cform.crud import crud
from plone.app.z3cform import wysiwyg
from plone.z3cform import z2
import collective.singing.interfaces

import OFS.interfaces

from collective.dancing import MessageFactory as _
from collective.dancing import collector
from collective.dancing.browser import controlpanel
from collective.dancing.browser import query

class ManageCollectorsForm(crud.CrudForm):
    """Crud form for collectors.
    """
    update_schema = field.Fields(ICollector).select('title')
    view_schema = field.Fields(ICollector).select('title')

    @property
    def add_schema(self):
        return self.update_schema + field.Fields(
            schema.Choice(
            __name__='factory',
            title=_(u"Type"),
            vocabulary=zope.schema.vocabulary.SimpleVocabulary(
                [zope.schema.vocabulary.SimpleTerm(value=f, title=f.title)
                 for f in collector.standalone_collectors])
            ))
    
    def get_items(self):
        return [(ob.getId(), ob) for ob in self.context.objectValues()]

    def add(self, data):
        name = Products.CMFPlone.utils.normalizeString(
            data['title'].encode('utf-8'), encoding='utf-8')
        self.context[name] = data['factory'](
            name, data['title'])
        return self.context[name]

    def remove(self, (id, item)):
        self.context.manage_delObjects([id])

    def link(self, item, field):
        if field == 'title':
            return item.absolute_url()

class CollectorAdministrationView(BrowserView):
    __call__ = ViewPageTemplateFile('controlpanel.pt')

    label = _(u'Collector administration')
    back_link = controlpanel.back_to_controlpanel

    def contents(self):
        z2.switch_on(self)
        return ManageCollectorsForm(self.context, self.request)()

collector_fields = field.Fields(
    collective.singing.interfaces.ICollector).select('title', 'optional')

def heading(self):
    if self.label:
        return "<h%s>%s</h%s>" % (self.level, self.label, self.level)

def prefix(self):
    return self.__class__.__name__ + '-'.join(self.context.getPhysicalPath())

class EditTopicForm(subform.EditSubForm):
    """Edit a single collector.
    """
    component.adapts(Products.ATContentTypes.content.topic.ATTopic,
                     None,
                     z3c.form.interfaces.IEditForm)
    template = viewpagetemplatefile.ViewPageTemplateFile('subform.pt')

    fields = field.Fields(
        schema.TextLine(__name__='title', title=_(u"Title")))

    @property
    def css_class(self):
        return "subform subform-level-%s" % self.level

    @property
    def label(self):
        return u"Collection: %s" % self.context.title

    prefix = property(prefix)

    def contents_bottom(self):
        return u'<a href="%s/criterion_edit_form">Edit the Smart Folder</a>' % (
            self.context.absolute_url())

    heading = heading

class EditTextForm(subform.EditSubForm):
    component.adapts(collector.ITextCollector,
                     None,
                     z3c.form.interfaces.IEditForm)
    template = viewpagetemplatefile.ViewPageTemplateFile('subform.pt')

    fields = z3c.form.field.Fields(collector.ITextCollector).select(
        'title', 'value')
    fields['value'].widgetFactory[
        z3c.form.interfaces.INPUT_MODE] = wysiwyg.WysiwygFieldWidget

    @property
    def css_class(self):
        return "subform subform-level-%s" % self.level

    @property
    def label(self):
        return u"Rich text: %s" % self.context.title

    prefix = property(prefix)

class EditReferenceForm(subform.EditSubForm):
    component.adapts(collector.IReferenceCollector,
                     None,
                     z3c.form.interfaces.IEditForm)
    template = viewpagetemplatefile.ViewPageTemplateFile('subform.pt')

    fields = z3c.form.field.Fields(
        collector.IReferenceCollector,
        query.IReferenceSelection).select('title', 'items')

    fields['items'].widgetFactory[
        z3c.form.interfaces.INPUT_MODE] = \
        z3c.formwidget.query.widget.QuerySourceFieldCheckboxWidget

    @property
    def css_class(self):
        return "subform subform-level-%s" % self.level

    @property
    def label(self):
        return u"Rich text: %s" % self.context.title

    prefix = property(prefix)

class AddToCollectorForm(form.Form):
    ignoreContext = True
    ignoreRequest = True
    template = viewpagetemplatefile.ViewPageTemplateFile('subform.pt')
    css_class = "addform"

    prefix = property(prefix)
    heading = heading

    @property
    def label(self):
        return u"Add item to %s" % self.context.title

    @property
    def fields(self):
        factory = schema.Choice(
            __name__='factory',
            title=_(u"Type"),
            vocabulary=zope.schema.vocabulary.SimpleVocabulary(
                [zope.schema.vocabulary.SimpleTerm(value=f, title=f.title)
                 for f in collector.collectors])
            )

        title = schema.TextLine(
            __name__='title',
            title=_(u"Title"))

        return z3c.form.field.Fields(factory, title)

    @z3c.form.button.buttonAndHandler(_('Add'), name='add')
    def handle_add(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = z3c.form.form.EditForm.formErrorsMessage
            return
        obj = data['factory'](self.context.get_next_id(), data['title'])
        self.context[obj.id] = obj
        self.status = _(u"Item added successfully.")

class DeleteFromCollectorForm(form.Form):
    template = viewpagetemplatefile.ViewPageTemplateFile('subform.pt')
    css_class = "deleteform"

    prefix = property(prefix)

    @z3c.form.button.buttonAndHandler(_('Remove block'), name='remove')
    def handle_remove(self, action):
        self.context.aq_parent.manage_delObjects([self.context.id])
        self.status = _("Item successfully deleted.")

class MoveBlockForm(form.Form):
    template = viewpagetemplatefile.ViewPageTemplateFile('subform.pt')
    css_class = "moveform"

    prefix = property(prefix)

    def _info_idx(self):
        infos = list(self.context.aq_parent._objects)
        info = None
        for info in infos:
            if info['id'] == self.context.id:
                break
        if info in infos:
            return infos.index(info)
        else:
            return 0

    def _move(self, delta):
        prev_index = self._info_idx()
        parent = self.context.aq_parent
        infos = list(parent._objects)
        my_info = infos[prev_index]
        del infos[prev_index]
        infos.insert(prev_index + delta, my_info)
        parent._objects = tuple(infos)

    @z3c.form.button.buttonAndHandler(
        _('Move block up'), name='up',
        condition=lambda form: form._info_idx() > 0)
    def handle_moveup(self, action):
        self._move(-1)
        self.status = _("Item successfully moved.")

    @z3c.form.button.buttonAndHandler(
        _('Move block down'), name='down',
        condition=lambda form: (form._info_idx() <
                                len(form.context.aq_parent._objects) - 1))
    def handle_moveup(self, action):
        self._move(1)
        self.status = _("Item successfully moved.")

class AbstractEditCollectorForm(object):
    level = 1
    
    @property
    def css_class(self):
        return "subform subform-level-%s" % self.level

    heading = heading

    def update(self):
        super(AbstractEditCollectorForm, self).update()
        addform = AddToCollectorForm(self.context, self.request)
        addform.level = self.level
        addform.update()

        actions = []
        for item in self.context.objectValues():
            acts = []
            for factory in DeleteFromCollectorForm, MoveBlockForm:
                form = factory(item, self.request)
                form.update()
                form.level = self.level + 1
                acts.append(form)
            actions.append(acts)

        editforms = []
        for item in self.context.objectValues():
             subform = component.getMultiAdapter(
                (item, self.request, self.parent_form),
                z3c.form.interfaces.ISubForm)
             subform.update()
             subform.level = self.level + 1
             editforms.append(subform)

        self.subforms = []
        for editform, acts in zip(editforms, actions):
            self.subforms.append(editform)
            self.subforms.extend(acts)
        self.subforms.append(addform)

class EditCollectorForm(AbstractEditCollectorForm, subform.EditSubForm):
    """Edit a single collector.
    """
    component.adapts(collective.singing.interfaces.ICollector,
                     zope.publisher.interfaces.http.IHTTPRequest,
                     z3c.form.interfaces.IEditForm)
    template = viewpagetemplatefile.ViewPageTemplateFile(
        'form-with-subforms.pt')
    fields = collector_fields

    prefix = property(prefix)

    @property
    def label(self):
        return u"Collector block: %s" % self.context.title

    @property
    def parent_form(self):
        return self.parentForm

class EditRootCollectorForm(AbstractEditCollectorForm, form.EditForm):
    """Edit a single collector.
    """
    template = viewpagetemplatefile.ViewPageTemplateFile(
        'form-with-subforms.pt')
    fields = collector_fields

    @property
    def parent_form(self):
        return self

class EditFilteredSubjectsCollectorForm(form.EditForm):
    template = viewpagetemplatefile.ViewPageTemplateFile('form.pt')

    @property
    def fields(self):
        schema = self.context.full_schema
        field = schema[self.context.field_name]
        field.__name__ = 'filtered_items'
        field.interface = None
        return z3c.form.field.Fields(schema)

class CollectorEditView(BrowserView):
    __call__ = ViewPageTemplateFile('controlpanel.pt')

    def label(self):
        return _(u'Edit ${collector}',
                 mapping=dict(collector=self.context.title))

    def back_link(self):
        return dict(label=_(u"Up to Collector administration"),
                    url=self.context.aq_inner.aq_parent.absolute_url())

    def contents(self):
        z2.switch_on(self)
        form = self.form(self.context, self.request)
        return '<div class="collector-form">' + form() + '</div>'

class RootCollectorEditView(CollectorEditView):
    form = EditRootCollectorForm

class FilteredCollectorEditView(CollectorEditView):
    form = EditFilteredSubjectsCollectorForm



import os
import tempfile

from zope.interface import Interface
import zc.lockfile
from Products.Five import BrowserView
from collective.singing.channel import channel_lookup
from collective.dancing import utils

LOCKFILE_NAME = os.path.join(tempfile.gettempdir(),
                             __name__ + '.tick_and_dispatch')

class IDancingUtilsView(Interface):
    
    def tick_and_dispatch(self):
        """Tick all schedulers of all channels.  Then dispatch their
        queues.
            
        This is what you call from cron or zope clock server to get
        periodic sending.
        """

    def handle_bounce(self):
        """Process a list of bouncing e-mail addresses.

        Expects a list of e-mail addresses in the ``addrs`` request
        variable.
        """

class DancingUtilsView(BrowserView):

    bounce_limit = 2

    def tick_and_dispatch(self):
        """ """
        try:
            lock = zc.lockfile.LockFile(LOCKFILE_NAME)
        except zc.lockfile.LockError:
            return "`tick_and_dispatch` is locked by another process (%r)." % (
                LOCKFILE_NAME)

        try:
            return self._tick_and_dispatch()
        finally:
            lock.close()

    def _tick_and_dispatch(self):
        msg = u''

        queue = utils.get_queue()
        num = queue.process()
        if num:
            for job in queue.finished[-num:]:
                msg += job.value + '\n'

        for channel in channel_lookup():
            queued = status = None
            if channel.scheduler is not None:
                queued = channel.scheduler.tick(channel, self.request)
            if channel.queue is not None:
                status = channel.queue.dispatch()
            d = {'channel':channel.name,
                 'queued':queued or 0,
                 'status':str(status or (0,0))}
            msg += u'%(channel)s: %(queued)d messages queued, dispatched: %(status)s\n' % d 
        return msg

    def handle_bounce(self):
        """ """
        count = 0
        addrs = self.request['addrs']
        if isinstance(addrs, (str, unicode)):
            addrs = [addrs]
        for channel in channel_lookup(only_subscribeable=True):
            for addr in addrs:
                subscriptions = channel.subscriptions.query(key=addr)
                for sub in subscriptions:
                    md = sub.metadata
                    bounces = md.get('bounces', 0)
                    if bounces >= self.bounce_limit:
                        md['pending'] = True
                        bounces = 0
                        count += 1
                    else:
                        bounces += 1
                    md['bounces'] = bounces
        return "%d addresses received, %d deactivated" % (len(addrs), count)



import z3c.form
from plone.z3cform import z2

from collective.dancing.browser.channel import ChannelAdministrationView
from collective.dancing.channel import INewslettersSettings

class NewslettersSettingsForm(z3c.form.form.EditForm):
    fields = z3c.form.field.Fields(INewslettersSettings)

class NewslettersSettingsView(ChannelAdministrationView):

    def contents(self):
        # Calling 'switch_on' is required before we can render z3c.forms.
        z2.switch_on(self)
        return NewslettersSettingsForm(self.context, self.request)()

